"""
Utilities for geopandas and other geographic information system tools
"""
import ubelt as ub
import numpy as np
import json
import os


def plot_geo_background():
    import kwplot
    import geopandas as gpd
    kwplot.autompl()
    # Broken in geopandas 1.x
    wld_map_gdf = gpd.read_file(
        gpd.datasets.get_path('naturalearth_lowres')
    ).to_crs('OGC:CRS84')
    ax = wld_map_gdf.plot()
    return ax


def geopandas_pairwise_overlaps(gdf1, gdf2, locator='iloc', predicate='intersects'):
    """
    Find pairwise relationships between each geometries

    Args:
        gdf1 (GeoDataFrame): query geo data

        gdf2 (GeoDataFrame): database geo data (builds spatial index)

        locator (str):
            either iloc for positional indexes or loc for pandas indexes.

        predicate (str, default='intersects'): a DE-9IM [1] predicate.
           (e.g. if intersection finds intersections between geometries)

    References:
        .. [DE_9IM] https://en.wikipedia.org/wiki/DE-9IM

    TODO:
        - [ ] This can move to kwgis.utils

    Returns:
        dict:
            mapping from integer-iloc-indexes in gdf1 to
            overlapping integer-iloc-indexes in gdf2

    Example:
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import geopandas as gpd
        >>> import geodatasets
        >>> geodatasets.get_path('GeoDa AirBnB')
        >>> gdf1 = gpd.read_file(geodatasets.get_path('GeoDa milwaukee1'))
        >>> gdf2 = gpd.read_file(geodatasets.get_path('GeoDa milwaukee2'))
        >>> #gdf1 = gpd.read_file(
        >>> #    gpd.datasets.get_path('naturalearth_lowres')
        >>> #)
        >>> #gdf2 = gpd.read_file(
        >>> #    gpd.datasets.get_path('naturalearth_cities')
        >>> #)
        >>> #gdf1 = gdf1.set_index('name')
        >>> #gdf2 = gdf2.set_index('name')
        >>> idx_mapping = geopandas_pairwise_overlaps(gdf1, gdf2, locator='iloc')
        >>> id_mapping = geopandas_pairwise_overlaps(gdf1, gdf2, locator='loc')

    Benchmark:
        import timerit
        ti = timerit.Timerit(10, bestof=3, verbose=2)
        for timer in ti.reset('with sindex O(N * log(M))'):
            with timer:
                fast_mapping = geopandas_pairwise_overlaps(gdf1, gdf2)
        for timer in ti.reset('without sindex O(N * M)'):
            with timer:
                from collections import defaultdict
                slow_mapping = defaultdict(list)
                for idx1, geom1 in enumerate(gdf1.geometry):
                    slow_mapping[idx1] = []
                    for idx2, geom2 in enumerate(gdf2.geometry):
                        if geom1.intersects(geom2):
                            slow_mapping[idx1].append(idx2)
        # check they are the same
        assert set(slow_mapping) == set(fast_mapping)
        for idx1 in slow_mapping:
            slow_idx2s = slow_mapping[idx1]
            fast_idx2s = fast_mapping[idx1]
            assert sorted(fast_idx2s) == sorted(slow_idx2s)
    """
    assert gdf1.index.is_unique, 'GeoDataFrame indexes must be unique'
    assert gdf2.index.is_unique, 'GeoDataFrame indexes must be unique'

    # Construct the spatial index (requires pygeos and/or rtree)
    sindex2 = gdf2.sindex
    # For each query polygon, lookup intersecting polygons in the spatial index
    idx1_to_idxs2 = {}
    for idx1, (rowid, row1) in enumerate(gdf1.iterrows()):
        idxs2 = sindex2.query(row1.geometry, predicate=predicate)
        # Record result indexes that "match" given the geometric predicate
        idx1_to_idxs2[idx1] = idxs2

    if locator == 'iloc':
        return idx1_to_idxs2
    elif locator == 'loc':
        _idxs1 = list(idx1_to_idxs2.keys())
        _nest_idxs2 = list(idx1_to_idxs2.values())

        id1_to_ids2 = {}
        for idx1, idxs2 in idx1_to_idxs2.items():
            id1 = gdf1.index[idx1]
            ids2 = gdf2.index[idxs2]
            id1_to_ids2[id1] = ids2
        if 0:
            import kwarray
            _flat_idxs2 = np.concatenate(_nest_idxs2, axis=0)
            _flat_ids2 = gdf2.index[_flat_idxs2]
            indexer = kwarray.FlatIndexer.fromlist(_nest_idxs2)
            _a = np.arange(len(indexer))
            outer, inner = indexer.unravel(_a)
            _ids1 = gdf1.index[_idxs1]
            _ids2 = list(kwarray.group_items(_flat_ids2.values, outer).values())

            list(map(len, _ids2))[0:10], list(map(len, _nest_idxs2))[0:10]
            list(map(len, _ids2))[-10:], list(map(len, _nest_idxs2))[-10:]
            list(map(len, _ids2))[0:10] == list(map(len, _nest_idxs2))[0:10]
            list(map(len, _ids2))[-10:] == list(map(len, _nest_idxs2))[-10:]
            np.vstack([[list(map(len, _ids2))], [list(map(len, _nest_idxs2))]])
            list(map(len, _ids2)) == list(map(len, _nest_idxs2))
            _id1_to_ids2 = ub.dzip(_ids1, _ids2)
            assert _id1_to_ids2 == id1_to_ids2
        return id1_to_ids2
    else:
        raise KeyError(locator)

    return idx1_to_idxs2


def latlon_text(lat, lon, precision=6):
    """
    Make a lat,lon string suitable for a filename.

    Pads with leading zeros so file names will align nicely at the same level
    of prcision.

    Args:
        lat (float): degrees latitude

        lon (float): degrees longitude

        precision (float, default=6):
            Number of trailing decimal places. As rule of thumb set this to:
                6 - for ~10cm accuracy,
                5 - for ~1m accuracy,
                2 - for ~1km accuracy,

    TODO:
        - [ ] This can move to kwgis.utils

    Notes:
        1 degree of latitude is *very* roughly the order of 100km, so the
        default precision of 6 localizes down to ~0.1 meters, which will
        usually be sufficient for satellite applications, but be mindful of
        using this text in applications that require more precision. Note 1
        degree of longitude will vary, but will always be at least as precise
        as 1 degree of latitude.

    Example:
        >>> lat = 90
        >>> lon = 180
        >>> print(latlon_text(lat, lon))
        N90.000000E180.000000

        >>> lat = 0
        >>> lon = 0
        >>> print(latlon_text(lat, lon))
        N00.000000E000.000000

    Example:
        >>> print(latlon_text(80.123, 170.123))
        >>> print(latlon_text(10.123, 80.123))
        >>> print(latlon_text(0.123, 0.123))
        N80.123000E170.123000
        N10.123000E080.123000
        N00.123000E000.123000

        >>> print(latlon_text(80.123, 170.123, precision=2))
        >>> print(latlon_text(10.123, 80.123, precision=2))
        >>> print(latlon_text(0.123, 0.123, precision=2))
        N80.12E170.12
        N10.12E080.12
        N00.12E000.12

        >>> print(latlon_text(80.123, 170.123, precision=5))
        >>> print(latlon_text(10.123, 80.123, precision=5))
        >>> print(latlon_text(0.123, 0.123, precision=5))
        N80.12300E170.12300
        N10.12300E080.12300
        N00.12300E000.12300
    """
    def _build_float_precision_fmt(num_leading, num_trailing):
        num2 = num_trailing
        # 2 extra for radix and leading sign
        num1 = num_leading + num_trailing + 2
        fmtparts = ['{:+0', str(num1), '.', str(num2), 'F}']
        fmtstr = ''.join(fmtparts)
        return fmtstr

    assert -90 <= lat <= 90, 'invalid lat'
    assert -180 <= lon <= 180, 'invalid lon'

    # Ensure latitude had 2 leading places and longitude has 3
    latfmt = _build_float_precision_fmt(2, precision)
    lonfmt = _build_float_precision_fmt(3, precision)

    lat_str = latfmt.format(lat).replace('+', 'N').replace('-', 'S')
    lon_str = lonfmt.format(lon).replace('+', 'E').replace('-', 'W')
    text = lat_str + lon_str
    return text


def demo_regions_geojson_text():
    import ubelt as ub
    geojson_text = ub.codeblock(
        '''
        {
            "type": "FeatureCollection",
            "features": [
                {
                    "type": "Feature",
                    "properties": {"type": "region", "region_model_id": "US_Jacksonville_R01", "version": "1.0.1", "mgrs": "17RMP", "start_date": "2009-05-09", "end_date": "2020-01-26" },
                    "geometry": {"type": "Polygon", "coordinates": [[[-81.6953, 30.3652], [-81.6942, 30.2984], [-81.5975, 30.2992], [-81.5968, 30.3667], [-81.6953, 30.3652]]]}
                },
                {
                    "type": "Feature",
                    "properties": {"type": "site", "site_id": "17RMP_US_Jacksonville_R01_0000", "start_date": "2016-02-14", "end_date": "2017-11-01"},
                    "geometry": {"type": "Polygon", "coordinates": [[[-81.6364, 30.3209], [-81.6364, 30.3236], [-81.6397, 30.3236], [-81.6397, 30.3209], [-81.6364, 30.3209]]]}
                },
                {
                    "type": "Feature",
                    "properties": {"type": "site", "site_id": "17RMP_US_Jacksonville_R01_0001", "start_date": "2016-07-13", "end_date": "2020-01-26" },
                    "geometry": {"type": "Polygon", "coordinates": [[[-81.6085, 30.3568], [-81.6085, 30.3600], [-81.6120, 30.3600], [-81.6120, 30.3568], [-81.6085, 30.3568]]]}
                }
            ]
        }
        ''')
    return geojson_text


def load_geojson(file, default_axis_mapping='OAMS_TRADITIONAL_GIS_ORDER'):
    """
    Args:
        file (str | file): path or file object containing geojson data.

        axis_mapping (str, default='OAMS_TRADITIONAL_GIS_ORDER'):
            The axis-ordering of the geojson file on disk.  This is assumed to
            be traditional ordering by default according to the geojson spec.

    Returns:
        GeoDataFrame : a dataframe with geo info.
            Note: geopandas always stores data in traditional XY, although its
            CRS does seem to hold axis order info?

        # OLD: This will ALWAYS return
        # with an OAMS_AUTHORITY_COMPLIANT wgs84 crs (i.e. lat,lon) even
        # though the on disk order is should be OAMS_TRADITIONAL_GIS_ORDER.

    References:
        https://geopandas.org/docs/user_guide/projections.html#the-axis-order-of-a-crs

    Example:
        >>> import io
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> geojson_text = demo_regions_geojson_text()
        >>> file = io.StringIO()
        >>> file.write(geojson_text)
        >>> file.seek(0)
        >>> region_df = load_geojson(file)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> import geopandas as gpd
        >>> kwplot.autompl()
        >>> wld_map_gdf = gpd.read_file(
        >>>     gpd.datasets.get_path('naturalearth_lowres')
        >>> ).to_crs('OGC:CRS84')
        >>> ax = wld_map_gdf.plot()
        >>> region_df.plot(ax=ax, edgecolor='orange', alpha=0.8)
        >>> # https://gis.stackexchange.com/questions/372564/userwarning-when-trying-to-get-centroid-from-a-polygon-geopandas
        >>> centroid = region_df.to_crs('+proj=cea').centroid.to_crs(region_df.crs)
        >>> centroid.plot(ax=ax, edgecolor='orange', alpha=0.8)

        print('region_df.crs = {!r}'.format(region_df.crs))
        print('wld_map_gdf.crs = {!r}'.format(wld_map_gdf.crs))
    """
    import geopandas as gpd
    valid_axis_mappings = {
        'OAMS_TRADITIONAL_GIS_ORDER',
        'OAMS_AUTHORITY_COMPLIANT',
    }
    if default_axis_mapping not in valid_axis_mappings:
        raise Exception

    # Read custom ROI regions
    region_df = gpd.read_file(file)

    # TODO: can we construct a pyproj.CRS from wgs84, but with traditional
    # order?

    # import pyproj
    # wgs84 = pyproj.CRS.from_epsg(4326)
    # z = pyproj.Transformer.from_crs(4326, 4326, always_xy=True)
    # crs1 = region_df.crs
    # pyproj.CRS.from_dict(crs1.to_json_dict())
    # z = region_df.crs
    # z.to_json_dict()

    # Use a CRS that actually reflects the underlying data
    if default_axis_mapping == 'OAMS_TRADITIONAL_GIS_ORDER':
        crs84 = get_crs84()
        # this is much faster and the only reason this is ok is because the
        # input is xy-wgs84 so the transform (which is slow) would be a noop
        region_df._crs = crs84
        # region_df = region_df.to_crs(crs84)
    else:
        raise NotImplementedError('geopandas only deals with traditional lon/lat')
    return region_df


read_geojson = load_geojson  # backwards compat


@ub.memoize
def get_crs84():
    """
    Constructing the CRS84 is slow.
    This function memoizes it so it only happens once.

    Returns:
        pyproj.crs.crs.CRS

    Example:
        >>> get_crs84()
    """
    from pyproj import CRS
    crs84 = CRS.from_user_input('OGC:CRS84')
    return crs84


_get_crs84 = get_crs84


def _flip(x, y):
    return (y, x)


def shapely_flip_xy(geom):
    from shapely import ops
    return ops.transform(_flip, geom)


def project_gdf_to_local_utm(gdf_crs84, max_utm_zones=1, mode=0,
                             tolerance=None):
    """
    Find the local UTM zone for a geo data frame and project to it.

    All geometry in the GDF must be in the same UTM zone.

    Args:
        gdf_crs84 (geopandas.GeoDataFrame):
            The data with CRS-84 geometry to project into a local UTM

        max_utm_zones (int):
            If the data spans more than this many UTM zones, error.
            Otherwise, we take the first one.

        mode (int):
            which version of the logic to use. Default is 0, which is the
            original.  Also have version 1 which is experimental and maybe
            faster.

        tolerance (int):
            if the projected region is further than this many meters from the
            center of the estimated UTM zone, raise a
            :class:`pyproj.exceptions.CRSError` error.
            Only used in mode 1.

    Returns:
        geopandas.GeoDataFrame:
            The a copy of the input dataframe with projected geometry.

    Notes:

        Assumes input geometry is in CRS-84.

        If an empty dataframe is given as input, it is returned as-is. No copy
        is made.


    Example:
        >>> import geopandas as gpd
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import kwarray
        >>> import kwimage
        >>> rng = kwarray.ensure_rng(0)
        >>> # Gen lat/lons between 0 and 1, which is in UTM zone 31N
        >>> gdf_crs84 = gpd.GeoDataFrame({'geometry': [
        >>>     kwimage.Polygon.random(rng=rng).to_shapely(),
        >>>     kwimage.Polygon.random(rng=rng).to_shapely(),
        >>>     kwimage.Polygon.random(rng=rng).to_shapely(),
        >>> ]}, crs='OGC:CRS84')
        >>> gdf_utm = project_gdf_to_local_utm(gdf_crs84)
        >>> assert gdf_utm.crs.name == 'WGS 84 / UTM zone 31N'

    Example:
        >>> import geopandas as gpd
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import kwarray
        >>> import kwimage
        >>> # If the data is too big for a single UTM zone,
        >>> rng = kwarray.ensure_rng(0)
        >>> gdf_crs84 = gpd.GeoDataFrame({'geometry': [
        >>>     kwimage.Polygon.random(rng=rng).scale(90).to_shapely(),
        >>>     kwimage.Polygon.random(rng=rng).scale(90).to_shapely(),
        >>>     kwimage.Polygon.random(rng=rng).scale(90).to_shapely(),
        >>> ]}, crs='OGC:CRS84')
        >>> import pytest
        >>> with pytest.raises(ValueError):
        >>>     gdf_utm = project_gdf_to_local_utm(gdf_crs84)

    Example:
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import geopandas as gpd
        >>> import kwarray
        >>> import kwimage
        >>> # Data convers a lot of utm zones
        >>> rng = kwarray.ensure_rng(0)
        >>> gdf_crs84 = gpd.GeoDataFrame({'geometry': [
        >>>     kwimage.Polygon.random(rng=rng).scale(2).translate((-1, -1)).scale(0.001, about='centroid').scale((180, 90)).to_shapely()
        >>>     for _ in range(5)
        >>> ]}, crs='OGC:CRS84')
        >>> # Mode 1 uses a tolerance test instead
        >>> gdf_utm = project_gdf_to_local_utm(gdf_crs84, mode=1, tolerance=float('inf'))
        >>> import pytest
        >>> import pyproj
        >>> with pytest.raises(pyproj.exceptions.CRSError):
        >>>     gdf_utm = project_gdf_to_local_utm(gdf_crs84, mode=1, tolerance=10000)
        >>> # Mode 1 uses the bounds of the entire region

    # TODO: Gracefully handle cases where the UTM zones are different
    # but all neighbors. Find a good example of this.
    # Example:
    #     >>> import geopandas as gpd
    #     >>> from kwgis.utils.util_gis import *  # NOQA
    #     >>> import kwarray
    #     >>> # If the data is too big for a single UTM zone,
    #     >>> rng = kwarray.ensure_rng(0)
    #     >>> # Corner case in Madagascar where the extent isn't too big, but it
    #     >>> # spans multiple UTM zones
    #     >>> poly = kwimage.Polygon.coerce({
    #     >>>     "type": "Polygon",
    #     >>>     "coordinates": [
    #     >>>         [[-73.77200379967688, 42.864783745778894],
    #     >>>          [-73.77177715301514, 42.86412514733195],
    #     >>>          [-73.77110660076141, 42.8641654498268],
    #     >>>          [-73.77105563879013, 42.86423720786224],
    #     >>>          [-73.7710489332676, 42.864399400374786],
    #     >>>          [-73.77134531736374, 42.8649134986743],
    #     >>>          [ -73.77200379967688, 42.864783745778894]]]
    #     >>> })
    #     >>> gdf_crs84 = gpd.GeoDataFrame({
    #     >>>     'geometry': [poly.to_shapely()]}, crs='OGC:CRS84')
    """
    # if gdf_crs84.crs.name != 'WGS 84 (CRS84)':
    #     raise AssertionError('expected CRS-84 input')
    # try:
    #     gdf_crs84.estimate_utm_crs()
    # except Exception:

    if len(gdf_crs84) == 0:
        return gdf_crs84

    if mode == 0:
        epsg_zones = []
        for geom_crs84 in gdf_crs84.geometry:
            epsg_zone = find_local_meter_epsg_crs(geom_crs84)
            epsg_zones.append(epsg_zone)

        if not ub.allsame(epsg_zones):
            unique_utm = set(epsg_zones)
            if len(unique_utm) > max_utm_zones:
                raise ValueError(ub.paragraph(
                    '''
                    Input data spanned multiple UTM zones.
                    This is currently not allowed. {}
                    '''
                ).format(unique_utm))
        # TODO: if there are more than one is there a way to get "the best one?"
        epsg_zone = epsg_zones[0]
        gdf_utm = gdf_crs84.to_crs(epsg_zone)
    elif mode == 1:
        utm_crs = gdf_crs84.estimate_utm_crs()
        gdf_utm = gdf_crs84.to_crs(utm_crs)
        if tolerance is not None:
            # not sure what a good default is here.
            import pyproj
            # Quick check first, then expensive check
            max_dist1 = np.abs(gdf_utm.bounds.values).max()
            if max_dist1 > tolerance:
                for geom in gdf_utm.geometry:
                    xys = np.array(geom.exterior.xy).T
                    max_dist = np.abs(xys).max()
                    if max_dist > tolerance:
                        if 1:
                            epsg_zones = []
                            for geom_crs84 in gdf_crs84.geometry:
                                epsg_zone = find_local_meter_epsg_crs(geom_crs84)
                                epsg_zones.append(epsg_zone)
                            print(f'epsg_zones={epsg_zones}')
                        raise pyproj.exceptions.CRSError(f'Projection distance={max_dist} is greater than tolerance={tolerance}')
    else:
        raise NotImplementedError(mode)
    return gdf_utm


def utm_epsg_from_latlon(lat, lon):
    """
    Find a reasonable UTM CRS for a given lat / lon

    The purpose of this function is to get a reasonable CRS for computing
    distances in meters. If the region of interest is very large, this may not
    be valid.

    Args:
        lat (float): degrees in latitude
        lon (float): degrees in longitude

    Returns:
        int : the ESPG code of the UTM zone

    References:
        https://gis.stackexchange.com/questions/190198/how-to-get-appropriate-crs-for-a-position-specified-in-lat-lon-coordinates
        https://gis.stackexchange.com/questions/365584/convert-utm-zone-into-epsg-code

    Example:
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> epsg_code = utm_epsg_from_latlon(0, 0)
        >>> print('epsg_code = {!r}'.format(epsg_code))
        epsg_code = 32631
    """
    import utm
    # easting, northing, zone_num, zone_code = utm.from_latlon(min_lat, min_lon)
    zone_num = utm.latlon_to_zone_number(lat, lon)

    # Construction of EPSG code from UTM zone number
    south = lat < 0
    epsg_code = 32600
    epsg_code += int(zone_num)
    if south is True:
        epsg_code += 100
    return epsg_code


class UTM_TransformContext:
    """
    Helper to project into UTM space, perform some transform, and then project
    back to the original CRS.

    Currently only supports CRS84

    Example:
        >>> import kwimage
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> data_crs84 = kwimage.Polygon.random()
        >>> with UTM_TransformContext(data_crs84) as self:
        >>>     orig_utm_poly = kwimage.Polygon.coerce(self.geoms_utm.iloc[0])
        >>>     new_utm_poly = orig_utm_poly.scale(2, about='center')
        >>>     self.finalize(new_utm_poly)
        >>> final_result = kwimage.Polygon.coerce(self.final_geoms_crs84.iloc[0])
        >>> naive_result = data_crs84.scale(2, about='center')
        >>> # Note the subtle difference in the naive vs context result
        >>> print(f'final_result={final_result}')
        >>> print(f'naive_result={naive_result}')
    """

    def __init__(self, data_crs84):
        """
        Args:
            data_crs84 (Coercible[GeoSeries]):
                something we know how to transform into a GeoSeries
        """
        from kwgis.utils import util_gis
        self.crs84 = util_gis.get_crs84()
        self.geoms_crs84 = self._coerce_geo_series(data_crs84, self.crs84)
        self.crs_utm = None
        self.gdf_utm = None
        self.final_geoms_utm = None
        self.final_geoms_crs84 = None

    def __enter__(self):
        from kwgis.utils import util_gis
        import geopandas as gpd
        gdf_crs84 = gpd.GeoDataFrame(geometry=self.geoms_crs84)
        gdf_utm = util_gis.project_gdf_to_local_utm(gdf_crs84)
        self.geoms_utm = gdf_utm.geometry
        self.crs_utm = self.geoms_utm.crs
        return self

    def _coerce_geo_series(self, data, default_crs=None):
        import shapely
        import geopandas as gpd
        import kwimage
        if isinstance(data, list):
            geoms = gpd.GeoSeries(data, crs=default_crs)
        else:
            if isinstance(data, shapely.geometry.base.BaseGeometry):
                geoms = gpd.GeoSeries([data], crs=default_crs)
            elif isinstance(data, kwimage.Polygon):
                geoms = gpd.GeoSeries([data.to_shapely()], crs=default_crs)
            elif isinstance(data, gpd.GeoDataFrame):
                geoms = data.geometry
            elif isinstance(data, gpd.GeoSeries):
                geoms = data
            else:
                raise TypeError(type(data))
        return geoms

    def finalize(self, final_utm):
        """
        Args:
            final_utm (Coercible[GeoSeries]):
                something coercable to geometry in UTM coordinates
        """
        final_geoms_utm = self._coerce_geo_series(
            final_utm, default_crs=self.crs_utm)
        self.final_geoms_utm = final_geoms_utm

    def __exit__(self, a, b, c):
        if self.final_geoms_utm is None:
            raise RuntimeError('Need to call finalize')
        self.final_geoms_crs84 = self.final_geoms_utm.to_crs(self.crs84)


def _demo_convert_latlon_to_utm():
    # Pretend we have these lon/lat (CRS84 corners)
    lat_y = 40.060759
    lon_x = 116.613095
    lat_y_off = 0.0001
    lat_x_off = 0.0001
    # hard code so north is up
    wld_corners = np.array([
        [lon_x - lat_x_off, lat_y + lat_y_off],
        [lon_x - lat_x_off, lat_y - lat_y_off],
        [lon_x + lat_x_off, lat_y - lat_y_off],
        [lon_x + lat_x_off, lat_y + lat_y_off],
    ])
    import kwimage
    wld_poly = kwimage.Polygon(exterior=wld_corners)
    wld_poly_sh = wld_poly.to_shapely()

    lon = wld_poly_sh.centroid.x
    lat = wld_poly_sh.centroid.y
    utm_code = utm_epsg_from_latlon(lat, lon)

    import geopandas as gpd
    gdf_crs84 = gpd.GeoDataFrame({'geometry': [wld_poly_sh]}, crs='OGC:CRS84')
    gdf_utm = gdf_crs84.to_crs(utm_code)
    utm_poly_sh = gdf_utm.iloc[0]['geometry']

    utm_poly = kwimage.Polygon.from_shapely(utm_poly_sh)
    utm_corners = utm_poly.data['exterior'].data

    min_x = utm_corners.T[0].min()
    max_x = utm_corners.T[0].max()
    min_y = utm_corners.T[1].min()
    max_y = utm_corners.T[1].max()

    # Note: UTM bottom should be the min value, so be careful here

    def _torch_meshgrid(*basis_dims):
        """
        References:
            https://zhaoyu.li/post/how-to-implement-meshgrid-in-pytorch/
        """
        basis_lens = list(map(len, basis_dims))
        new_dims = []
        for i, basis in enumerate(basis_dims):
            # Probably a more efficent way to do this, but its right
            newshape = [1] * len(basis_dims)
            reps = list(basis_lens)
            newshape[i] = -1
            reps[i] = 1
            dd = basis.view(*newshape).repeat(*reps)
            new_dims.append(dd)
        return new_dims

    import torch
    image_width = 256
    image_height = 256

    utm_x_basis = torch.linspace(min_x, max_x, image_width)
    utm_y_basis = torch.linspace(max_y, min_y, image_height)

    xgrid, ygrid = _torch_meshgrid(utm_x_basis, utm_y_basis)


def find_local_meter_epsg_crs(geom_crs84):
    """
    Find the "best" meter based CRS for a smallish geographic region.

    Currently this only returns UTM zones. Might be better to return an Albers
    projection if the geometry spans more than one UTM zone.

    Args:
        geom_crs84 (shapely.geometry.base.BaseGeometry):
            shapely geometry in CRS84 (lon/lat wgs84)

    Returns:
        int: epsg code

    References:
        [1] https://gis.stackexchange.com/questions/148181/choosing-projection-crs-for-short-distance-based-analysis/148187
        [2] http://projfinder.com/

    Example:
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import kwimage
        >>> geom_crs84 = kwimage.Polygon.random().translate(-0.5).scale((180, 90)).to_shapely()
        >>> epsg_zone = find_local_meter_epsg_crs(geom_crs84)

    SeeAlso:
        GeoDataFrame.estimate_utm_crs

    TODO:
        - [ ] Albers?
        - [ ] Better UTM zone intersection
        - [ ] Fix edge cases
    """

    import geopandas as gpd
    utm_crs = gpd.array.from_shapely([geom_crs84], 'OGC:CRS84').estimate_utm_crs()
    epsg_zone = utm_crs.to_epsg()

    return epsg_zone


def check_latlons(lat, lon):
    """
    Quick check to see if latitudes and longitudes are valid.

    Longitude (x) is always between -180 and 180 (degrees east)
    Latitude (y) is always between -90 and 90 (degrees north)

    Example:
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import pytest
        >>> assert not check_latlons(1000, 1000)
        >>> assert check_latlons(0, 0)
    """
    lat = np.asarray(lat)
    lon = np.asarray(lon)
    bad_lat = (lat < -90).any() or (lat > 90).any()
    bad_lon = (lon < -180).any() or (lon > 180).any()
    in_ranges = not (bad_lon or bad_lat)
    return in_ranges


def coerce_geojson_datas(arg, format='dataframe', allow_raw=False, workers=0,
                         mode='thread', verbose=1, desc=None, parse_float=None):
    """
    Attempts to resolve an argument into multiple geojson datas.

    Multiple threads / processes are used to load the specified information and
    the function generates dictionaries of information containing the file path
    and the loaded data as they become available.

    The argument can be:

            1. A path to a geojson file (or a list of them)

            2. A glob string specifying multiple geojson files (or a list of them)

            3. A path to a json manifest file.

            4. If allow_raw is True, then the input can be a raw json string,
               dict, or GeoDataFrame.

    Args:
        arg (str | PathLike | List[str | PathLike]):
            an argument that is coerceable to one or more GeoDataFrames.

        format (str):
            Indicates the returned format of the data. Can be 'dataframe' where
            the 'data' key will be a GeoDataFrame, or 'json' where the raw json
            data will be returned.

        allow_raw (bool):
            if True, we will also check if the arguments are raw json /
            geopandas data that can be loaded. In general try not to enable
            this.

        workers (int):
            number of io workers

        mode (str):
            concurrent executor mode. Can be 'serial', 'thread', or 'process'.

        desc (str):
            custom message for progress bar.

    Yields:
        List[Dict[str, Any | GeoDataFrame | Dict]]:
            A list of dictionaries formated with the keys:

                * fpath (str): the file path the data was loaded from (
                    if applicable)

                * data (GeoDataFrame | dict):
                    the data loaded in the requested format

    SeeAlso:
        * load_site_or_region_dataframes - the function that does the loading
            after the arguments are coerced.

    Example:
        >>> # xdoctest: +SKIP("failing on CI. unsure why")
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> from kwgis.demo.metrics_demo import generate_demodata
        >>> info1 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R001')
        >>> info2 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R002')
        >>> info3 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R003')
        >>> info4 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R012')
        >>> info5 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R022')
        >>> #
        >>> region_fpaths = sorted(info1['true_region_dpath'].glob('*.geojson'))
        >>> site_fpaths = sorted(info1['true_site_dpath'].glob('*.geojson'))
        >>> #
        >>> import json
        >>> manifest_fpath1 =  info1['output_dpath'] / 'demo_manifest1.json'
        >>> manifest_fpath2 =  info1['output_dpath'] / 'demo_manifest2.json'
        >>> manifest_data1 = {
        >>>     'files': [str(p) for p in region_fpaths[0:2]]
        >>> }
        >>> manifest_data2 = {
        >>>     'files': [str(p) for p in region_fpaths[3:4]]
        >>> }
        >>> manifest_fpath1.write_text(json.dumps(manifest_data1))
        >>> manifest_fpath2.write_text(json.dumps(manifest_data2))
        >>> variants = []
        >>> # ==========
        >>> # Test Cases
        >>> # ==========
        >>> #
        >>> # List of region files
        >>> arg = region_fpaths
        >>> result = list(coerce_geojson_datas(arg))
        >>> assert len(result) == 5
        >>> #
        >>> # Glob for region files
        >>> arg = str(info1['true_region_dpath']) + '/*R*2*.geojson'
        >>> result = list(coerce_geojson_datas(arg))
        >>> assert len(result) == 3
        >>> #
        >>> # Manifest file
        >>> arg = manifest_fpath1
        >>> result = list(coerce_geojson_datas(arg))
        >>> assert len(result) == 2
        >>> #
        >>> # Manifest file glob
        >>> arg = str(manifest_fpath1.parent / '*.json')
        >>> result = list(coerce_geojson_datas(arg))
        >>> assert len(result) == 3
        >>> #
        >>> # Manifest file glob and a region path
        >>> arg = [manifest_fpath2, region_fpaths[0]]
        >>> result = list(coerce_geojson_datas(arg))
        >>> assert len(result) == 2
        >>> #
        >>> # Site glob and a manifest glob
        >>> arg = [str(info1['true_site_dpath']) + '/DR_R002_*.geojson',
        ...        str(manifest_fpath1 + '*')]
        >>> result = list(coerce_geojson_datas(arg))
        >>> assert len(result) == 9
        >>> #
        >>> # Site directory and manifest file.
        >>> arg = [str(info1['true_site_dpath']),
        ...        str(manifest_fpath1 + '*')]
        >>> result = list(coerce_geojson_datas(arg))
        >>> assert len(result) == 31

        >>> # Test raw loading and format swapping
        >>> from kwgis.utils import util_gis
        >>> arg = util_gis.demo_regions_geojson_text()
        >>> result1 = list(coerce_geojson_datas(arg, allow_raw=False))
        >>> assert len(result1) == 0
        >>> result2 = list(coerce_geojson_datas(arg, allow_raw=True))
        >>> assert len(result2) == 1
        >>> arg = result2
        >>> result3 = list(coerce_geojson_datas(arg, format='dataframe', allow_raw=True))
        >>> assert result3 == result2
        >>> result4 = list(coerce_geojson_datas(arg, format='json', allow_raw=True))
        >>> assert isinstance(result4[0]['data'], dict)
        >>> result5 = list(coerce_geojson_datas(
        >>>     result4, format='dataframe', allow_raw=True))
        >>> assert isinstance(result4[0]['data'], dict)

        >>> #
        >>> # Test nothing case
        >>> assert len(list(coerce_geojson_datas([], allow_raw=True))) == 0
    """
    if format not in {'json', 'dataframe'}:
        raise KeyError(format)

    def is_iterable_sequence(item, strok=False):
        """
        Check if it is a non-string, non-mapping type of iterable.
        """
        if ub.iterable(item, strok=strok):
            if hasattr(item, 'keys'):
                return False
            else:
                return True
        else:
            return False

    if arg is None:
        return

    if allow_raw:
        # Normally the function assumes we are only inputing things that are
        # coercable to paths, and then to geojson. But sometimes we might want
        # to pass around that data directly. In this case, grab those items
        # first, and then resolve the rest of them.

        if is_iterable_sequence(arg):
            iterable_arg = arg
        else:
            iterable_arg = [arg]

        raw_items = []
        other_items = []
        for item in iterable_arg:
            try:
                was_raw, item = _coerce_raw_geojson(item, format)
            except TypeError:
                print(f'iterable_arg={iterable_arg}')
                print(f'arg={arg}')
                raise
            if was_raw:
                raw_items.append(item)
            else:
                other_items.append(item)
        path_coercable = other_items
    else:
        path_coercable = arg

    # Handle the normal better-defined case of coercing arguments into paths
    geojson_fpaths = coerce_geojson_paths(path_coercable)

    if allow_raw:
        if verbose:
            if raw_items or len(geojson_fpaths) == 0:
                print(f'Coerced {len(raw_items)} raw geojson item')
            if raw_items:
                if len(geojson_fpaths) == 0:
                    # Disable path verbosity if there were raw items, but no
                    # paths.
                    verbose = 0

    # Now all of resolved accumulator items should be geojson files.
    # Submit the data to be loaded.
    geojson_fpaths = list(ub.unique(geojson_fpaths))
    data_gen = load_geojson_datas(
        geojson_fpaths, workers=workers, mode=mode, desc=desc,
        format=format, verbose=verbose, yield_after_submit=True,
        parse_float=parse_float)

    # Start the background workers
    next(data_gen)

    if allow_raw:
        # yield the raw data before the generated data
        yield from raw_items

    # Finish the main generator
    yield from data_gen


def coerce_geojson_paths(data, return_manifests=False):
    """
    Resolves the argument to a list of geojson paths.

    Args:
        data (Any): argument to coerce.
            The argument can be a full path, a glob string, a path to a
            manifest file or any combination of the previous in a list.
            If it is a manifest file. It must be a json/yaml file with a
            top-level "files" key that points to a list of the actual file
            paths.

        return_manifests (bool): if True additionally returns paths to
            any intermediate manifest files.

    Example:
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import json
        >>> from kwgis.demo.metrics_demo import generate_demodata
        >>> # Setup a bunch of geojson files
        >>> outdir = ub.Path.appdir("kwgis/tests/gis/coerce_geojson_v1")
        >>> # outdir.delete().ensuredir()
        >>> info1 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R001', outdir=outdir)
        >>> info2 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R002', outdir=outdir)
        >>> info3 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R003', outdir=outdir)
        >>> info4 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R012', outdir=outdir)
        >>> info5 = generate_demodata.generate_demo_metrics_framework_data(roi='DR_R022', outdir=outdir)
        >>> region_fpaths = sorted(info1['true_region_dpath'].glob('*.geojson'))
        >>> site_fpaths = sorted(info1['true_site_dpath'].glob('*.geojson'))
        >>> manifest_fpath1 =  info1['output_dpath'] / 'demo_manifest1.json'
        >>> manifest_data1 = {'files': [str(p) for p in region_fpaths[0:2]]}
        >>> geojson_dpath = info1['true_site_dpath']
        >>> manifest_fpath1.write_text(json.dumps(manifest_data1))
        >>> # Test manifest case
        >>> geojson_fpaths = coerce_geojson_paths(manifest_fpath1)
        >>> assert len(geojson_fpaths) == 2
        >>> # Test directory case
        >>> geojson_fpaths = coerce_geojson_paths(geojson_dpath)
        >>> assert len(geojson_fpaths) == 15, 'maybe cache is bad?'
        >>> # Test glob case
        >>> geojson_fpaths = coerce_geojson_paths(geojson_dpath / '*R001_*')
        >>> assert len(geojson_fpaths) == 3
        >>> # Test list of files and globstr
        >>> data = [geojson_dpath / '*R002_*'] + geojson_fpaths
        >>> geojson_fpaths = coerce_geojson_paths(data)
        >>> assert len(geojson_fpaths) == 6
        >>> # Test manifest case2
        >>> info = coerce_geojson_paths(manifest_fpath1, return_manifests=True)
        >>> assert len(info['manifest_fpaths']) == 1
        >>> assert len(info['geojson_fpaths']) == 2
    """
    from kwutil import util_yaml
    from kwutil import util_path
    paths = util_path.coerce_patterned_paths(data, '.geojson')
    geojson_fpaths = []
    manifest_fpaths = []
    for p in paths:
        resolved = None
        if isinstance(p, (str, os.PathLike)):
            peeked = None
            if str(p).lower().endswith('.json'):
                # Check to see if this is a manifest file
                peeked = json.loads(p.read_text())
            elif str(p).lower().endswith(('.yaml', '.yml')):
                # Check to see if this is a manifest file
                peeked = util_yaml.Yaml.loads(p.read_text())
            if peeked is not None:
                if isinstance(peeked, dict) and 'files' in peeked:
                    manifest_fpaths.append(p)
                    resolved = list(map(ub.Path, peeked['files']))
        if resolved is None:
            resolved = [p]

        if resolved is None:
            raise ValueError('Unable to resolve p={p!r}')
        geojson_fpaths.extend(resolved)

    if return_manifests:
        return {
            'manifest_fpaths': manifest_fpaths,
            'geojson_fpaths': geojson_fpaths,
        }
    else:
        return geojson_fpaths


def _coerce_raw_geojson(item, format):
    """
    Helper for the coerce method
    """
    import geopandas as gpd
    was_raw = False

    if isinstance(item, dict):
        # Allow the item to be a wrapped dict returned by this func
        was_raw = True
        if set(item.keys()) == {'fpath', 'data', 'format'}:
            item = item['data']

    if isinstance(item, str):
        # Allow the item to be unparsed
        try:
            item = json.loads(item)
        except json.JSONDecodeError:
            ...  # not json data
        else:
            was_raw = True

    if isinstance(item, (os.PathLike, str)):
        ...  # not raw
    elif isinstance(item, dict):
        # Allow the item to be parsed json
        was_raw = True
        assert item.get('type', None) == 'FeatureCollection'
        if format == 'dataframe':
            item = gpd.GeoDataFrame.from_features(item['features'])
            # Hack in CRS-84
            crs84 = get_crs84()
            # this is much faster and the only reason this is ok is because the
            # input is xy-wgs84 so the transform (which is slow) would be a noop
            item._crs = crs84
    elif isinstance(item, gpd.GeoDataFrame):
        # Allow the item to be a GeoDataFrame
        was_raw = True
        if format == 'json':
            item = json.loads(item.to_json())
    else:
        raise TypeError(type(item))
    if was_raw:
        item = {
            'fpath': None,
            'data': item,
            'format': format,
        }
    return was_raw, item


def _load_json_from_path(path, parse_float=None):
    with open(path, 'r') as file:
        return json.load(file, parse_float=parse_float)


def load_geojson_datas(geojson_fpaths, format='dataframe', workers=0,
                       mode='thread', verbose=1, desc=None,
                       yield_after_submit=False,
                       parse_float=None):
    """
    Generator that loads sites (and the path they loaded from) in parallel

    Args:
        geojson_fpaths (Iterable[PathLike]):
            geojson paths to load

        format (str):
            Can be "dataframe" (for pandas) or "json" (for nested dict / list)

        workers (int):
            number of background loading workers

        mode (str):
            concurrent executor mode

        desc (str): overwrite message for the progress bar

        yield_after_submit (bool):
            backend argument that will yield None after the data is submitted
            to force the data loading to start processing in the background.

    TODO:
        pass something in to allow for better progress management when a lot of
        functions are calling this.

    Yields:
        Dict:
            containing keys, 'fpath' and 'gdf'.

    SeeAlso:
        * coerce_geojson_datas - the coercable version of this function.
        * coerce_geojson_paths - only coerces paths

    Example:
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import ubelt as ub
        >>> dpath = ub.Path.appdir('kwgis', 'tests', 'util_gis', 'load_geojson_data')
        >>> dpath.ensuredir()
        >>> fpath = dpath / 'data.geojson'
        >>> fpath.write_text(demo_regions_geojson_text())
        >>> # Test both format loaders work correctly.
        >>> gdf = list(load_geojson_datas([fpath], format='dataframe'))[0]['data']
        >>> dct = list(load_geojson_datas([fpath], format='json'))[0]['data']
        >>> import geopandas as gpd
        >>> assert isinstance(gdf, gpd.GeoDataFrame)
        >>> assert isinstance(dct, dict)
    """
    from kwgis.utils import util_gis
    from kwutil import util_progress
    # sites = []
    if desc is None:
        desc = 'load geojson datas'

    jobs = ub.JobPool(mode=mode, max_workers=workers)
    submit_progkw = {
        'desc': 'submit ' + desc,
        'verbose': (workers > 0) and verbose
    }

    # FIXME: kwutil.util_progress is not properly disabling progress when
    # verbose=0, fix that and then reenable this.
    # try:
    #     num_jobs = len(geojson_fpaths)
    # except Exception:
    #     num_jobs = None

    # if num_jobs is not None and num_jobs < 1000:
    #     # Don't bother with a progress bar for submit jobs
    #     # if there are not too many of them.
    #     submit_progkw['verbose'] = 0
    #     submit_progkw['enabled'] = False

    kwargs = {}
    if format == 'dataframe':
        loader = util_gis.load_geojson
    elif format == 'json':
        loader = _load_json_from_path
        kwargs['parse_float'] = parse_float
    else:
        raise KeyError(format)

    pmankw = {}
    if submit_progkw['verbose'] == 0:
        # Hack because rich based progiters always cause a newline
        # switching to progiter backend prevents this.
        # TODO: fix this issue in util_progress.ProgressManager itself
        pmankw['backend'] = 'progiter'

    pman = util_progress.ProgressManager(**pmankw)
    with pman:

        for fpath in pman.progiter(geojson_fpaths, **submit_progkw):
            job = jobs.submit(loader, fpath, **kwargs)
            job.fpath = fpath

        if yield_after_submit:
            yield None

        result_progkw = {
            'verbose': verbose,
        }
        for job in pman.progiter(jobs.as_completed(), total=len(jobs),
                                 desc=desc, **result_progkw):
            data = job.result()
            info = {
                'fpath': job.fpath,
                'data': data,
                'format': format,
            }
            yield info


def crs_geojson_to_gdf(geometry, crs_info=None):
    """
    TODO: fix the name and scope of this function.

    The idea is to convert one of our CRS aware geojson geometries to a
    GeoDataFrame for manipulation.

    Args:
        geometry :
            This is a geojson geometry object with that has a crs_info object
            in its properties.

        crs_info (Dict | None):
            if the geojson does not have a crs_info in its properties you
            can specify it here.

    Returns:
        gpd.GeoDataFrame

    Example:
        >>> # xdoctest: +REQUIRES(env:LARGE_DOWNLOAD)
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> import kwimage
        >>> from kwgis.demo.dummy_demodata import dummy_rpc_geotiff_fpath
        >>> gpath_or_ref = gpath = dummy_rpc_geotiff_fpath()
        >>> info = geotiff_crs_info(gpath)
        >>> # Case 1: Traditional CRS84 Geojson
        >>> geometry = info['geos_corners']
        >>> gdf = crs_geojson_to_gdf(geometry)
        >>> assert gdf.crs.axis_info[0].name == 'Geodetic longitude'
        >>> assert gdf.crs.axis_info[1].name == 'Geodetic latitude'
        >>> # Case 2: Authority Compliant WGS84 Geojson
        >>> wgs84_poly = kwimage.Polygon.coerce(info['wgs84_corners'])
        >>> wgs84_geometry = wgs84_poly.to_geojson()
        >>> wgs84_geometry['properties'] = {'crs_info': info['wgs84_crs_info']}
        >>> crs84_gdf = crs_geojson_to_gdf(wgs84_geometry)
        >>> assert crs84_gdf.crs.axis_info[0].name == 'Geodetic longitude'
        >>> assert crs84_gdf.crs.axis_info[1].name == 'Geodetic latitude'
        >>> # Case 3: UTM
        >>> utm_poly = kwimage.Polygon.coerce(info['utm_corners'])
        >>> utm_geometry = utm_poly.to_geojson()
        >>> utm_geometry['properties'] = {'crs_info': info['utm_crs_info']}
        >>> utm_gdf = crs_geojson_to_gdf(utm_geometry)
        >>> assert utm_gdf.crs.axis_info[0].name == 'Easting'
        >>> assert utm_gdf.crs.axis_info[1].name == 'Northing'
    """
    import kwimage
    import geopandas as gpd
    geos_crs_info = geometry['properties'].get('crs_info', None)
    if geos_crs_info is None:
        geos_crs_info = crs_info
    if geos_crs_info is None:
        raise Exception("we need crs_info")
    # TODO: better way to get generic Polygon / Multipolygon from shaley in
    # kwimage.
    kw_geom = kwimage.structs.segmentation._coerce_coco_segmentation(geometry)
    # sh_geom = kwimage.MultiPolygon.from_geojson(geometry).to_shapely()
    auth = geos_crs_info['auth']
    if isinstance(auth, tuple):
        auth = list(auth)
    if auth == ['EPSG', '4326']:
        # TODO:
        # - [ ] Handle general axis_mapping issues
        if geos_crs_info['axis_mapping'] != 'OAMS_TRADITIONAL_GIS_ORDER':
            if geos_crs_info['axis_mapping'] == 'OAMS_AUTHORITY_COMPLIANT':
                kw_geom = kw_geom.swap_axes()
            else:
                raise NotImplementedError(
                    'geojson should be in traditional order. i.e. crs84'
                )
        crs = get_crs84()
    else:
        from pyproj import CRS
        crs = CRS.from_user_input(geos_crs_info['auth'])
        if geos_crs_info['axis_mapping'] != 'OAMS_AUTHORITY_COMPLIANT':
            raise NotImplementedError(
                f'Non-CRS84 should be authority compliant. Got: {geos_crs_info}'
            )
    sh_geom = kw_geom.to_shapely()
    gdf = gpd.GeoDataFrame({'geometry': [sh_geom]}, crs=crs)
    return gdf


def coerce_crs(crs_info):
    """
    Get a pyproj CRS from something they should be inferrable from

    Example:
        >>> from kwgis.gis.geotiff import *  # NOQA
        >>> from kwgis.utils.util_gis import *  # NOQA
        >>> coerce_crs('crs84')
        >>> coerce_crs(['EPSG', '4326'])
    """
    if isinstance(crs_info, str):
        if crs_info == 'crs84':
            crs = get_crs84()
        else:
            from pyproj import CRS
            crs = CRS.from_user_input(crs_info['auth'])
    elif isinstance(crs_info, dict):
        auth = crs_info['auth']
        if isinstance(auth, tuple):
            auth = list(auth)
        # if auth == ['EPSG', '4326']:
        #     if crs_info['axis_mapping'] != 'OAMS_TRADITIONAL_GIS_ORDER':
        #         raise ValueError(
        #             'got real wgs crs. Should be in traditional order. i.e. crs84'
        #         )
        # else:
        from pyproj import CRS
        crs = CRS.from_user_input(crs_info['auth'])
        # if crs_info['axis_mapping'] != 'OAMS_AUTHORITY_COMPLIANT':
        #     raise NotImplementedError(
        #         f'Non-CRS84 should be authority compliant. Got: {crs_info}'
        #     )
    else:
        from pyproj import CRS
        crs = CRS.from_user_input(crs_info)
    # TODO: detect wgs84 and error, must assume crs84.
    return crs


def utm_epsg_from_zoneletter(zone, letter):
    """
    Example:
        >>> lon, lat = -99.1149386165299831, 19.4867151641771414
        >>> espg1 = utm_epsg_from_latlon(lat, lon)
        >>> zone = 14
        >>> letter = 'Q'
        >>> espg2 = utm_epsg_from_zoneletter(zone, letter)
        >>> assert espg2 == espg1

    References:
        https://www.maptools.com/tutorials/grid_zone_details

    Notes:
        The UTM coordinate system divides the earth into 60 zones each 6
        degrees of longitude wide. These zones define the reference point for
        UTM grid coordinates within the zone. UTM zones extend from a latitude
        of 80° S to 84° N. In the polar regions the Universal Polar
        Stereographic (UPS) grid system is used. Note that there are a few
        exceptions to zone width in Northern Europe to keep small countries in
        a single zone.

        UTM zones are numbered 1 through 60, starting at the international date
        line, longitude 180°, and proceeding east. Zone 1 extends from 180° W
        to 174° W and is centered on 177° W.

        Each zone is divided into horizontal bands spanning 8 degrees of
        latitude. These bands are lettered, south to north, beginning at 80° S
        with the letter C and ending with the letter X at 84° N. The letters I
        and O are skipped to avoid confusion with the numbers one and zero. The
        band lettered X spans 12° of latitude.
    """
    if letter.upper() >= 'N':
        epsg_code = 32600 + int(zone)  # Northern hemisphere
    else:
        epsg_code = 32700 + int(zone)  # Southern hemisphere
    return epsg_code



import numpy as np
from dateutil.parser import isoparse


def group_tiles(query, timediff_sec=300):
    """
    Given a RGDC search query result, nest results so that scenes
    that can be merged are grouped together

    TODO when RGDC STAC endpoint is available, calculate overlap with AOI here

    Args:
        query: output of rgdc.search
        timediff_sec: max allowed time between adjacent scenes.
            Should nominally be 23 seconds for Landsat 7 and 8. Some buffer is included.
            TODO what is it for Sentinel?
    Example:
        >>> # xdoctest: +SKIP
        >>> from rgd_client import Rgdc
        >>> client = Rgdc('username', 'password')
        >>> query = (client.search(**kwargs, instrumentation='ETM') +
        >>>          client.search(**kwargs, instrumentation='OLI_TIRS'))
        >>> # query == [scene1, scene2, scene3, scene4]
        >>> query = group_tiles(query)
        >>> # query == [[scene1], [scene2, scene3], [scene4]]
    """
    # ensure we're only working with one satellite at a time
    query = sorted(query, key=lambda q: isoparse(q['acquisition_date']))
    sensors, ixs = np.unique([q['instrumentation'] for q in query],
                             return_index=True)
    assert set(sensors).issubset({'ETM', 'OLI_TIRS', 'S2A', 'S2B'}), f'unsupported sensor in {sensors}'
    to_process = np.split(query, ixs[1:])

    def _path_row(q):
        """
        Returns LS path and row as strings. eg ('115', '034')
        """
        path_row = q['subentry_name'].split('_')[2]
        path, row = path_row[:3], path_row[3:]
        return path, row

    result = []
    for sensor, query in zip(sensors, to_process):
        # extract and split by datetimes
        dts = [isoparse(q['acquisition_date']) for q in query]
        diffs = np.diff([dt.timestamp() for dt in dts],
                        prepend=dts[0].timestamp())
        ixs = np.where(diffs > timediff_sec)[0]

        for split in np.split(query, ixs):
            # for Landsat, an additional check:
            # each split should consist of adjacent rows in the same path.
            if sensor in {'ETM', 'OLI_TIRS'}:
                path_rows = [_path_row(s) for s in split]
                paths, rows = zip(*path_rows)
                assert len(np.unique(paths)) == 1
                assert len(np.unique(rows)) == len(split)

            result.append(split)

    return result


def bands_landsat(entry):
    """
    Get only the band files from a Landsat RasterMetaEntry downloaded from RGD.

    Args:
        entry: RasterMetaEntry, the output type of rgdc.download_raster

    Returns:
        list of pathlib paths to band files
    """
    return [str(p) for p in entry.images if 'BQA' not in p.stem]


def bands_sentinel2(entry):
    """
    Get only the band files from a Sentinel-2 RasterMetaEntry downloaded from RGD.

    Args:
        entry: RasterMetaEntry, the output type of rgdc.download_raster

    Returns:
        list of pathlib paths to band files
    """
    return [str(p) for p in entry.images if p.match('*_B*.jp2')]



"""
A submodule where each file contains information and functionality specific to
a particular GEO sensor.
"""

__dev__ = """
mkinit ~/code/watch/kwgis/gis/sensors/__init__.py -w
"""
from kwgis.gis.sensors import sentinel2

from kwgis.gis.sensors.sentinel2 import (sentinel2_grid,)

__all__ = ['sentinel2', 'sentinel2_grid']



"""
Collected information about satellite bands from https://github.com/stac-extensions/eo/

This should be mostly independent of the data source used.
(eg Google Cloud vs element84 AWS, Collection-1 vs Collection-2)
No guarantees that every value will be indentical, but close enough for heuristics.

Each constant is a list of 'eo:bands' Band Object dicts.
They are sorted in the same order in which they appear if the bands come stacked in a single image,
or in lexicographic order if the bands come in separate images.

Coverage of the catalogs is inconsistent. Where necessary, info has been filled in by hand.


Notes:

    Sentinal 2 Band Table
    =====================
    Band    Resolution    Central Wavelength    Description
    B1            60 m                443 nm    Ultra blue (Coastal and Aerosol)
    B2            10 m                490 nm    Blue
    B3            10 m                560 nm    Green
    B4            10 m                665 nm    Red
    B5            20 m                705 nm    Visible and Near Infrared (VNIR)
    B6            20 m                740 nm    Visible and Near Infrared (VNIR)
    B7            20 m                783 nm    Visible and Near Infrared (VNIR)
    B8            10 m                842 nm    Visible and Near Infrared (VNIR)
    B8a           20 m                865 nm    Visible and Near Infrared (VNIR)
    B9            60 m                940 nm    Short Wave Infrared (SWIR)
    B10           60 m               1375 nm    Short Wave Infrared (SWIR)
    B11           20 m               1610 nm    Short Wave Infrared (SWIR)
    B12           20 m               2190 nm    Short Wave Infrared (SWIR)


    Landsat 8 Band Table
    =====================
    Band    Resolution    Central Wavelength    Description
    1            30 m                 430 nm    Coastal aerosol
    2            30 m                 450 nm    Blue
    3            30 m                 530 nm    Green
    4            30 m                 640 nm    Red
    5            30 m                 850 nm    Near Infrared (NIR)
    6            30 m                1570 nm    SWIR 1
    7            30 m                2110 nm    SWIR 2
    8            15 m                 500 nm    Panchromatic
    9            30 m                1360 nm    Cirrus
    10           100 m              10600 nm    Thermal Infrared (TIRS) 1
    11           100 m              11500 nm    Thermal Infrared (TIRS) 2


    Worldview 3 MUL Band Table
    ==========================
    Band    Resolution    Central Wavelength    Description
    1           1.38 m                 400 nm    Coastal aerosol
    2           1.38 m                 450 nm    Blue
    3           1.38 m                 510 nm    Green
    4           1.38 m                 585 nm    Yellow
    5           1.38 m                 630 nm    Red
    6           1.38 m                 705 nm    Red edge
    7           1.38 m                 770 nm    Near-IR1
    8           1.38 m                 860 nm    Near-IR2

    Worldview 3 PAN Band Table
    ==========================
    1           0.34 m                 450-800 nm  Panchromatic

References:
    https://gis.stackexchange.com/questions/290796/how-to-edit-the-metadata-for-individual-bands-of-a-multiband-raster-preferably
    https://gisgeography.com/sentinel-2-bands-combinations/
    https://earth.esa.int/eogateway/missions/worldview-3
    https://www.usgs.gov/faqs/what-are-band-designations-landsat-satellites?qt-news_science_products=0#qt-news_science_products
    .. [MODIS-SPEC] https://modis.gsfc.nasa.gov/about/specifications.php
    .. [SentinelHub] https://apps.sentinel-hub.com/eo-browser/?zoom=15&lat=42.87425&lng=-73.83164&themeId=DEFAULT-THEME&visualizationUrl=https%3A%2F%2Fservices.sentinel-hub.com%2Fogc%2Fwms%2Fbd86bcc0-f318-402b-a145-015f85b9427e&datasetId=S2L2A&fromTime=2022-02-16T00%3A00%3A00.000Z&toTime=2022-02-16T23%3A59%3A59.999Z&layerId=4-FALSE-COLOR-URBAN
"""


def dicts_contain(d_list, dsub_list):
    def contains(ds):
        return all(ds[0][k] == ds[1][k] for k in ds[1])
    return all(map(contains, zip(d_list, dsub_list)))


'''
This band info is taken from the sentinelhub AWS catalog.
It will need to be updated to match RGD's when that is STAC-compliant.

Note this is for S2B; S2A has slightly different center_wavelength and full_width_half_max

References:
    https://www.element84.com/earth-search/
    https://docs.sentinel-hub.com/api/latest/data/sentinel-2-l1c

Example:
    >>> from pystac_client import Client
    >>> from kwgis.utils.util_bands import *
    >>> cat = Client.open('https://earth-search.aws.element84.com/v1')
    >>> search = cat.search(bbox=[-110, 39.5, -105, 40.5], max_items=1, collections=['sentinel-2-l1c'])
    >>> i = list(search.items())[0]
    >>> # one image per band
    >>> bands = [v.to_dict()['eo:bands'][0] for k,v in i.assets.items() if k.startswith('B')]
    >>> assert dicts_contain(SENTINEL2, bands)
'''
SENTINEL2 = [
    {'name': 'B01', 'common_name': 'coastal', 'gsd': 60,
     'center_wavelength': 0.4439, 'full_width_half_max': 0.027},
    {'name': 'B02', 'common_name': 'blue', 'gsd': 10,
     'center_wavelength': 0.4966, 'full_width_half_max': 0.098},
    {'name': 'B03', 'common_name': 'green', 'gsd': 10,
     'center_wavelength': 0.56, 'full_width_half_max': 0.045},
    {'name': 'B04', 'common_name': 'red', 'gsd': 10,
     'center_wavelength': 0.6645, 'full_width_half_max': 0.038},
    {'name': 'B05',
     'gsd': 20,
     'center_wavelength': 0.7039,
     'full_width_half_max': 0.019},
    {'name': 'B06',
     'gsd': 20,
     'center_wavelength': 0.7402,
     'full_width_half_max': 0.018},
    {'name': 'B07',
     'gsd': 20,
     'center_wavelength': 0.7825,
     'full_width_half_max': 0.028},
    {'name': 'B08', 'gsd': 10, 'common_name': 'nir',
     'center_wavelength': 0.8351, 'full_width_half_max': 0.145},
    {'name': 'B8A',
     'gsd': 20,
     'center_wavelength': 0.8648,
     'full_width_half_max': 0.033},
    {'name': 'B09',
     'gsd': 60,
     'center_wavelength': 0.945,
     'full_width_half_max': 0.026},
    {'name': 'B10', 'gsd': 60, 'common_name': 'cirrus',
     'center_wavelength': 1.3735, 'full_width_half_max': 0.075},
    {'name': 'B11', 'gsd': 20, 'common_name': 'swir16',
     'center_wavelength': 1.6137, 'full_width_half_max': 0.143},
    {'name': 'B12', 'gsd': 20, 'common_name': 'swir22',
     'center_wavelength': 2.22024, 'full_width_half_max': 0.242}
]


'''
This band info is taken from the sentinelhub AWS catalog.
It will need to be updated to match RGD's when that is STAC-compliant.

References:
    https://www.element84.com/earth-search/
    https://docs.sentinel-hub.com/api/latest/data/landsat-8/
    https://landsat.gsfc.nasa.gov/satellites/landsat-8/
    https://planetarycomputer.microsoft.com/dataset/landsat-c2-l2

Example:
    >>> from pystac_client import Client
    >>> # for Collection 1
    >>> cat = Client.open('https://earth-search.aws.element84.com/v0')
    >>> search = cat.search(bbox=[-110, 39.5, -105, 40.5], max_items=1, collections=['landsat-8-l1-c1'])
    >>> # for Collection 2
    >>> # cat = Client.open('https://earth-search.aws.element84.com/v1')
    >>> # search = cat.search(bbox=[-110, 39.5, -105, 40.5], max_items=1, collections=['landsat-ot-l1'])
    >>> i = list(search.items())[0]
    >>> # one image for all bands
    >>> bands = [v.to_dict()['eo:bands'][0] for k,v in i.assets.items() if k.startswith('B') and (k != 'BQA')]
    >>>
    >>> from kwgis.utils.util_bands import *
    >>> assert dicts_contain(LANDSAT8, bands)
'''
LANDSAT8 = [
    {'name': 'B1', 'common_name': 'coastal', 'gsd': 30,
     'center_wavelength': 0.48, 'full_width_half_max': 0.02,
     'alias': ['aerosol']},

    {'name': 'B2', 'common_name': 'blue', 'gsd': 30,
     'center_wavelength': 0.44, 'full_width_half_max': 0.06},

    {'name': 'B3', 'common_name': 'green', 'gsd': 30,
     'center_wavelength': 0.56, 'full_width_half_max': 0.06},

    {'name': 'B4', 'common_name': 'red', 'gsd': 30,
     'center_wavelength': 0.65, 'full_width_half_max': 0.04},

    {'name': 'B5', 'common_name': 'nir', 'gsd': 30,
     'center_wavelength': 0.86, 'full_width_half_max': 0.03,
     'alias': ['nir08']},

    {'name': 'B6', 'common_name': 'swir16', 'gsd': 30,
     'center_wavelength': 1.6, 'full_width_half_max': 0.08},

    {'name': 'B7', 'common_name': 'swir22', 'gsd': 30,
     'center_wavelength': 2.2, 'full_width_half_max': 0.2},

    {'name': 'B8', 'common_name': 'pan', 'gsd': 15,
     'center_wavelength': 0.59, 'full_width_half_max': 0.18},

    {'name': 'B9', 'common_name': 'cirrus', 'gsd': 30,
     'center_wavelength': 1.37, 'full_width_half_max': 0.02},

    {'name': 'B10', 'common_name': 'lwir11', 'gsd': 100,
     'center_wavelength': 10.9, 'full_width_half_max': 0.8,
     'alias': ['tir1'], 'notes': 'thermal-ir'},

    {'name': 'B11', 'common_name': 'lwir12', 'gsd': 100,
     'center_wavelength': 12, 'full_width_half_max': 1,
     'alias': ['tir2'], 'notes': 'thermal-ir'},
]


'''
This band info is taken from the USGS Landsat catalog.

This is for Collection-2 Level-1; may be slightly different from
Collection-1 Level-1 (RGD's current source)

Example:
    >>> # not compatible with pystac_client for some reason
    >>> import requests
    >>> item = requests.get(('https://landsatlook.usgs.gov/sat-api/collections'
    >>>                      '/landsat-c2l1/items/LE07_L1TP_026043_20210518_20210518_02_RT')).json()
    >>> assets = item['assets']
    >>> keys = sorted(k for k in assets.keys() if 'B' in k)
    >>> bands = [assets[k]['eo:bands'][0] for k in keys]
    >>>
    >>> from kwgis.utils.util_bands import *
    >>> assert dicts_contain(LANDSAT7, bands)
'''
LANDSAT7 = [
    {'name': 'B1', 'common_name': 'blue', 'gsd': 30, 'center_wavelength': 0.49},
    {'name': 'B2', 'common_name': 'green', 'gsd': 30, 'center_wavelength': 0.56},
    {'name': 'B3', 'common_name': 'red', 'gsd': 30, 'center_wavelength': 0.66},
    {'name': 'B4', 'common_name': 'nir08', 'gsd': 30, 'center_wavelength': 0.84},
    {'name': 'B5', 'common_name': 'swir16', 'gsd': 30, 'center_wavelength': 1.65},
    {'name': 'B6', 'common_name': 'tir', 'gsd': 30,
        'center_wavelength': 11.45},  # B6_VCID_1.TIF
    {'name': 'B6', 'common_name': 'tir', 'gsd': 30,
        'center_wavelength': 11.45},  # B6_VCID_2.TIF
    {'name': 'B7', 'common_name': 'swir22', 'gsd': 30, 'center_wavelength': 2.22},
    {'name': 'B8', 'common_name': 'pan', 'gsd': 30, 'center_wavelength': 0.71}
]


'''
This band info is taken from the IARPA T&E STAC catalog.

Example:
    >>> # xdoctest: +SKIP
    >>> # requires the api_key for this catalog
    >>> from pystac_client import Client
    >>> catalog = Client.open('https://api.smart-stac.com/', headers={"x-api-key": api_key})
    >>> search = catalog.search(collections=['worldview-nitf'], bbox=[128.662489, 37.659517, 128.676673, 37.664560])
    >>> items = list(search.items())
    >>> props = [i.to_dict()['properties'] for i in items]
    >>>
    >>> wv01 = [p for p in props if p['mission'] == 'WV01']
    >>> assert np.unique([p['instruments'] for p in wv01]) == ['panchromatic']
    >>> wv01_pan = [p for p in wv01 if p['instruments'] == ['panchromatic']][0]['eo:bands']
    >>>
    >>> wv02 = [p for p in props if p['mission'] == 'WV02']
    >>> assert np.all(np.unique([p['instruments'] for p in wv02]) == ['panchromatic', 'vis-multi'])
    >>> assert np.all(np.unique([len(p['eo:bands']) for p in wv02]) == [1, 4, 8])
    >>> wv02_pan = [p for p in wv02 if p['instruments'] == ['panchromatic']][0]['eo:bands']
    >>> wv02_ms = [p['eo:bands'] for p in wv02 if p['instruments'] == ['vis-multi']]
    >>> wv02_ms4 = [p for p in wv02_ms if len(p) == 4][0]
    >>> wv02_ms8 = [p for p in wv02_ms if len(p) == 8][0]
    >>>
    >>> wv03 = [p for p in props if p['mission'] == 'WV03']
    >>> assert np.all(np.unique([p['instruments'] for p in wv03]) == ['panchromatic', 'vis-multi'])
    >>> assert np.all(np.unique([len(p['eo:bands']) for p in wv03]) == [1, 8])
    >>> wv03_pan = [p for p in wv03 if p['instruments'] == ['panchromatic']][0]['eo:bands']
    >>> wv03_ms = [p['eo:bands'] for p in wv03 if p['instruments'] == ['vis-multi']]
    >>> wv03_ms8 = [p for p in wv03_ms if len(p) == 8][0]
    >>>
    >>> # not sure if this must be true, but it is
    >>> assert wv02_pan == wv03_pan
    >>> assert wv02_ms8 == wv03_ms8
    >>>
    >>> from kwgis.utils.util_bands import *
    >>> assert dicts_contain(WORLDVIEW1_PAN, wv01_pan)
    >>> assert dicts_contain(WORLDVIEW2_PAN, wv02_pan)
    >>> assert dicts_contain(WORLDVIEW2_MS4, wv02_ms4)
    >>> assert dicts_contain(WORLDVIEW2_MS8, wv02_ms8)
    >>> assert dicts_contain(WORLDVIEW3_PAN, wv03_pan)
    >>> assert dicts_contain(WORLDVIEW3_MS8, wv03_ms8)
'''
WORLDVIEW1_PAN = [
    {'name': 'PAN', 'common_name': 'panchromatic', 'center_wavelength': 0.65}
]

WORLDVIEW2_PAN = [
    {'name': 'PAN', 'common_name': 'panchromatic', 'center_wavelength': 0.625}
]
WORLDVIEW2_MS4 = [
    {'name': 'B2', 'common_name': 'blue', 'center_wavelength': 0.48},
    {'name': 'B3', 'common_name': 'green', 'center_wavelength': 0.545},
    {'name': 'B5', 'common_name': 'red', 'center_wavelength': 0.66},
    {'name': 'B7', 'common_name': 'near-ir1', 'center_wavelength': 0.833}
]
WORLDVIEW2_MS8 = [
    {'name': 'B1', 'common_name': 'coastal', 'center_wavelength': 0.425},
    {'name': 'B2', 'common_name': 'blue', 'center_wavelength': 0.48},
    {'name': 'B3', 'common_name': 'green', 'center_wavelength': 0.545},
    {'name': 'B4', 'common_name': 'yellow', 'center_wavelength': 0.605},
    {'name': 'B5', 'common_name': 'red', 'center_wavelength': 0.66},
    {'name': 'B6', 'common_name': 'red-edge', 'center_wavelength': 0.725},
    {'name': 'B7', 'common_name': 'near-ir1', 'center_wavelength': 0.833},
    {'name': 'B8', 'common_name': 'near-ir2', 'center_wavelength': 0.95}
]


WORLDVIEW3_PAN = [
    {'name': 'PAN', 'common_name': 'panchromatic', 'center_wavelength': 0.625}
]
# TODO does WORLDVIEW3_MS4 not exist or did we just happen to not get it
# in the initial catalog?
WORLDVIEW3_MS8 = [
    {'name': 'B1', 'common_name': 'coastal', 'center_wavelength': 0.425},
    {'name': 'B2', 'common_name': 'blue', 'center_wavelength': 0.48},
    {'name': 'B3', 'common_name': 'green', 'center_wavelength': 0.545},
    {'name': 'B4', 'common_name': 'yellow', 'center_wavelength': 0.605},
    {'name': 'B5', 'common_name': 'red', 'center_wavelength': 0.66},
    {'name': 'B6', 'common_name': 'red-edge', 'center_wavelength': 0.725},
    {'name': 'B7', 'common_name': 'near-ir1', 'center_wavelength': 0.833},
    {'name': 'B8', 'common_name': 'near-ir2', 'center_wavelength': 0.95}
]

PLANETSCOPE_3BAND = [
    {'name': 'B1', 'common_name': 'red'},
    {'name': 'B2', 'common_name': 'green'},
    {'name': 'B3', 'common_name': 'blue'}
]

PLANETSCOPE_4BAND = [
    {'name': 'B1', 'common_name': 'red'},
    {'name': 'B2', 'common_name': 'green'},
    {'name': 'B3', 'common_name': 'blue'},
    {'name': 'B4', 'common_name': 'nir'}
]

PLANETSCOPE_8BAND = [
    {'name': 'B1', 'common_name': 'coastal'},
    {'name': 'B2', 'common_name': 'blue'},
    {'name': 'B3', 'common_name': 'green1'},
    {'name': 'B4', 'common_name': 'green'},
    {'name': 'B5', 'common_name': 'yellow'},
    {'name': 'B6', 'common_name': 'red'},
    {'name': 'B7', 'common_name': 'red-edge'},
    {'name': 'B8', 'common_name': 'nir'}
]

'''
TODO

fix wv doctest
'''

ALL_BANDS = (
    SENTINEL2 + LANDSAT8 + LANDSAT7 + WORLDVIEW1_PAN + WORLDVIEW2_PAN +
    WORLDVIEW2_MS4 + WORLDVIEW2_MS8 + WORLDVIEW3_PAN + WORLDVIEW3_MS8)

'''
WIP
Collect synonyms for allowed common_names values (not enforced by STAC)
TODO do we even need to conform to this? Should we only collect
"true" synonyms like {'pan': 'panchromatic'} ?

Example:
    >>> from kwgis.utils.util_bands import *
    >>> import itertools
    >>> names = set(b.get('common_name', '') for b in ALL_BANDS)
    >>> accounted_names = set(EO_COMMONNAMES.keys()).union(
    >>>     set(itertools.chain.from_iterable(EO_COMMONNAMES.values())))
    >>> todo = names.difference(accounted_names)
    >>> # not sure what to do with these
    >>> print(todo)
    {'', 'tir'}

References:
    https://github.com/stac-extensions/eo/blob/main/json-schema/schema.json#L151
'''
EO_COMMONNAMES = {
    'coastal': [],
    'blue': [],
    'green': [],
    'red': [],
    'rededge': ['red-edge'],
    'yellow': [],
    'pan': ['panchromatic'],
    'nir': ['near-ir1', 'near-ir2'],
    'nir08': [],
    'nir09': [],
    'cirrus': [],
    'swir16': [],
    'swir22': [],
    'lwir': [],
    'lwir11': [],
    'lwir12': []
}


'''
WIP
Bands that are used to observe targets on the ground
This is just a rough first pass

Example:
    >>> from kwgis.utils.util_bands import *
    >>> assert GROUND.issubset(set(EO_COMMONNAMES.keys()))
'''
GROUND = {
    'coastal',
    'blue',
    'green',
    'red',
    'rededge',
    'yellow',
    'pan',
    'nir',
    'nir08',
    'nir09',
}

'''
These band fields can be accessed as python objects as well using pystac

Example:
    >>> from pystac.extensions.eo import Band
    >>> from kwgis.utils.util_bands import *
    >>> for band in ALL_BANDS:
    >>>     band.pop('gsd', None)  # pystac doesn't support this yet
    >>>     b = Band.create(**band)
'''


# https://modis.gsfc.nasa.gov/about/specifications.php
MODIS = {
    'name': 'MODIS',
    'quantization': '12 bits',
    'bands': [
        {'band': 1, 'gsd': '250m', 'bandwidth': '620-670nm', 'spectral_radiance': 21.8, 'required_snr': 128, 'primary_use': 'Land/Cloud/Aerosols Boundaries'},
        {'band': 2, 'gsd': '250m', 'bandwidth': '841-876nm', 'spectral_radiance': 24.7, 'required_snr': 201, 'primary_use': 'Land/Cloud/Aerosols Boundaries'},

        {'band': 3, 'gsd': '500m', 'bandwidth': '459 - 479nm', 'spectral_radiance': 35.3, 'required_snr': 243, 'primary_use': 'Land/Cloud/Aerosols Properties'},
        {'band': 5, 'gsd': '500m', 'bandwidth': '545 - 565nm', 'spectral_radiance': 29.0, 'required_snr': 228, 'primary_use': 'Land/Cloud/Aerosols Properties'},
        {'band': 5, 'gsd': '500m', 'bandwidth': '1230 - 1250nm', 'spectral_radiance': 5.4, 'required_snr': 74, 'primary_use': 'Land/Cloud/Aerosols Properties'},
        {'band': 6, 'gsd': '500m', 'bandwidth': '1628 - 1652nm', 'spectral_radiance': 7.3, 'required_snr': 275, 'primary_use': 'Land/Cloud/Aerosols Properties'},
        {'band': 7, 'gsd': '500m', 'bandwidth': '2105 - 2155nm', 'spectral_radiance': 1.0, 'required_snr': 110, 'primary_use': 'Land/Cloud/Aerosols Properties'},

        {'band': 8, 'gsd': '1000m', 'bandwidth': '405 - 420nm', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
        {'band': 9, 'gsd': '1000m', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
        {'band': 10, 'gsd': '1000m', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
        {'band': 11, 'gsd': '1000m', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
        {'band': 12, 'gsd': '1000m', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
        {'band': 13, 'gsd': '1000m', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
        {'band': 14, 'gsd': '1000m', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
        {'band': 15, 'gsd': '1000m', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
        {'band': 16, 'gsd': '1000m', 'bandwidth': '862 - 877nm', 'primary_use': 'Ocean Color/Phytoplankton/Biogeochemistry'},
    ]
}


# https://ladsweb.modaps.eosdis.nasa.gov/missions-and-measurements/viirs/
# https://lpdaac.usgs.gov/data/get-started-data/collection-overview/missions/s-npp-nasa-viirs-overview/
# https://rammb.cira.colostate.edu/projects/npp/VIIRS_bands_and_bandwidths.pdf
# The VIIRS instrument provides 22 spectral bands from 412 nanometers (nm) to 12
# micrometers (µm) at two spatial resolutions, 375 meters (m) and 750 m, which
# are resampled to 500 m, 1 km, and 0.05 degrees in the NASA produced data
# products to promote consistency with the MODIS heritage.
VIIRS = {
    'name': 'VIIRS',
    'quantization': '12 bits',
    'gds': '750m',
    'bands': [
        {'name': 'I1', 'gsd': '375m', 'reflected_range': '0.6 - 0.68um', 'desc': 'Near Infrared', 'common_name': 'blue'},
        {'name': 'I2', 'gsd': '375m', 'common_name': 'green'},
        {'name': 'I3', 'gsd': '375m', 'common_name': 'red'},
        {'name': 'I4', 'gsd': '375m'},
        {'name': 'I5', 'gsd': '375m'},

        {'name': 'M1', 'gsd': '750m'},
        {'name': 'M2', 'gsd': '750m'},
        {'name': 'M3', 'gsd': '750m'},
        {'name': 'M4', 'gsd': '750m'},
        {'name': 'M5', 'gsd': '750m'},
        {'name': 'M6', 'gsd': '750m'},
        {'name': 'M7', 'gsd': '750m'},
        {'name': 'M8', 'gsd': '750m'},
        {'name': 'M9', 'gsd': '750m'},
        {'name': 'M10', 'gsd': '750m'},
        {'name': 'M11', 'gsd': '750m'},
        {'name': 'M12', 'gsd': '750m'},
        {'name': 'M13', 'gsd': '750m'},
        {'name': 'M14', 'gsd': '750m'},
        {'name': 'M15', 'gsd': '750m'},
        {'name': 'M16', 'gsd': '750m', 'reflected_range': '11.54 - 12.49um'},

        {'name': 'DNB', 'gsd': '750m', 'reflected_range': '0.5 - 0.9um', 'desc': 'Visible/Reflective'},
    ]
}


SENTINEL1 = {
    'name': 'Sentinel1',
    'notes': 'SAR',
    'gsd': '1.7x4.3m - 3.6x4.9m',
}


# TODO: move to a new "Spectral-Index" module

def specialized_index_bands(bands=None, coco_img=None, symbolic=False):
    r"""
    Ported from code from by (Yongquan Zhao on 26 April 2017)

    References:
        https://mail.google.com/mail/u/1/#chat/space/AAAAE5jpxTc

    Ignore:
        DVC_DPATH=$HOME/data/dvc-repos/smart_watch_dvc
        jq '.images[0].id' $DVC_DPATH/drop1-S2-L8-aligned/data.kwcoco.json
        kwcoco subset --src $DVC_DPATH/drop1-S2-L8-aligned/data.kwcoco.json --gids=2, --dst=./one_image_data/data.kwcoco.json --copy_assets=True

    Ignore:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> from kwgis.utils.util_bands import *  # NOQA
        >>> from simple_dvc.registry import find_dvc_dpath
        >>> import kwcoco
        >>> import ubelt as ub
        >>> dvc_dpath = find_dvc_dpath()
        >>> coco_fpath = dvc_dpath / 'drop1-S2-L8-aligned/data.kwcoco.json'
        >>> dset = kwcoco.CocoDataset(coco_fpath)
        >>> from kwgis.utils import kwcoco_extensions
        >>> gid = ub.peek(dset.index.imgs.keys())
        >>> vidid = dset.index.name_to_video['BH_Manama_R01']['id']
        >>> gid = dset.index.vidid_to_gids[vidid][0]
        >>> coco_img = dset.coco_image(gid)
        >>> print('coco_img.channels = {!r}'.format(coco_img.channels))
        >>> symbolic = False
        >>> indexes = specialized_index_bands(coco_img=coco_img)
        >>> #indexes = ub.dict_isect(indexes, {"ASI", 'AF_Norm', 'SSF_Norm', 'VSF_Norm', 'MF_Norm'})
        >>> indexes = ub.dict_isect(indexes, {"ASI"})
        >>> import kwarray
        >>> print(ub.urepr(ub.map_vals(kwarray.stats_dict, indexes), nl=1))
        >>> import pandas as pd
        >>> print(pd.DataFrame(ub.map_vals(kwarray.stats_dict, indexes)))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> import kwimage
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(indexes))
        >>> #value = kwimage.normalize_intensity(coco_img.imdelay('red|green|blue').finalize())
        >>> #kwplot.imshow(value, title='red|green|blue', pnum=pnum_())
        >>> for key, value in indexes.items():
        >>>     kwplot.imshow(kwimage.normalize(value), title=key, pnum=pnum_())


    # Example:
    #     >>> # xdoctest: +REQUIRES(module:sympy)
    #     >>> from kwgis.utils.util_bands import *  # NOQA
    #     >>> symbolic = True
    #     >>> indexes = specialized_index_bands(coco_img=None, symbolic=symbolic)
    #     >>> import sympy as sym
    #     >>> for key, index in indexes.items():
    #     >>>     print('===============')
    #     >>>     print('key = {!r}'.format(key))
    #     >>>     print('\nOrig {}'.format(key))
    #     >>>     print(index)
    #     >>>     print('\nSimplified {}'.format(key))
    #     >>>     print(sym.simplify(index))
    """
    import ubelt as ub
    import numpy as np
    import kwimage

    if bands is not None:
        allbands = np.stack(list(ub.take(bands, ['blue', 'green', 'red', 'swir16', 'swir22', 'nir'])))
        allbands = kwimage.normalize_intensity(allbands)
        Blue, Green, Red, SWIR1, SWIR2, NIR = allbands

    # Raw bands
    elif symbolic:
        # Sympy can help explore different forms of these equations.
        import sympy as sym
        Blue, Green, Red, SWIR1, SWIR2, NIR = sym.symbols(
            'Blue, Green, Red, SWIR1, SWIR2, NIR')
    else:
        delayed = coco_img.imdelay()
        rgbir123 = delayed.take_channels('blue|green|red|swir16|swir22|nir')
        chw = rgbir123.finalize().transpose(2, 0, 1)

        chw = kwimage.normalize_intensity(chw)

        Blue = chw[0]    # NOQA
        Green = chw[1]  # NOQA
        Red = chw[2]      # NOQA
        SWIR1 = chw[3]  # NOQA
        SWIR2 = chw[4]  # NOQA
        NIR = chw[5]      # NOQA

    Blue = Blue     # NOQA
    Grn = Green   # NOQA
    Red = Red      # NOQA
    SWIR2 = SWIR1   # NOQA
    SWIR1 = SWIR2  # NOQA
    NIR = NIR     # NOQA

    def hist_cut(band, fill_value=0, k=1, minmax='std'):
        if minmax == 'std':
            mean = band.mean()
            std = band.std()
            low_val = (mean + k * std)
            high_val = (mean + k * std)
        else:
            low_val, high_val = minmax
        # is_low = band < low_val
        # is_high = band > high_val
        # [is_high] = fill
        band = band.clip(low_val, high_val)
        return band

    def minmax_norm(band, mask):
        mask = mask & np.isfinite(band)
        if mask.sum() > 0:
            max_val = band[mask].max()
            min_val = band[mask].min()
            extent = max_val - min_val
            if extent > 0:
                shifted = band - min_val
                scaled = shifted / extent
                band[mask] = scaled[mask]
        return band

    # TODO:
    # Need a valid_mask
    valid_mask = None
    if valid_mask is None:
        valid_mask = np.ones(Blue.shape, dtype=bool)

    MinMaxNorm = minmax_norm
    HistCut = hist_cut
    MaskValid = valid_mask

    # constants
    G = 2.5
    C1 = 6
    C2 = 7.5

    # L = 1
    L = 1000

    fillV = 0

    # Might be tricky
    # toplogical_effects = False

    # Formulas

    # Artificial surface Factor (AF).
    AF = ((SWIR1 + NIR) / 2 - Blue) / ((SWIR1 + NIR) / 2 + Blue)
    AF = HistCut(AF, fillV, 6, [-1, 1])
    AF_Norm = MinMaxNorm(AF, MaskValid)

    # Vegetation Suppressing Factor (VSF).
    # EVI is better in general cases, but its adjustment for mountain shadows
    # (with vegetation) is not as good as NDVI.
    EVI = G * ((NIR - Red) / (NIR + C1 * Red - C2 * Blue + L))
    EVI = HistCut(EVI, fillV, 6, [-1, 1])
    VSF = 1 - EVI
    # NDVI = (NIR - Red) / (NIR + Red)
    #  NDVI  = HistCut( NDVI, fillV, 6, [-1 1])
    # VSF = 1 - NDVI
    VSF_Norm = MinMaxNorm(VSF, MaskValid)

    # Soil Suppressing Factor (SSF).
    # Derive the Modified Bare soil Index (MBI).
    MBI = (SWIR1 - SWIR2 - NIR) / (SWIR1 + SWIR2 + NIR) + 0.5
    MBI = HistCut(MBI, fillV, 6, [-0.5, 1.5])
    MBI_Norm = MinMaxNorm(MBI, MaskValid)
    # Deriving Enhanced-MBI based on MBI and MNDWI.
    MNDWI = (Grn - SWIR1) / (Grn + SWIR1)
    MNDWI = HistCut(MNDWI, fillV, 6, [-1, 1])
    MNDWI_Norm = MinMaxNorm(MNDWI, MaskValid)
    EMBI = ((MBI_Norm) - (MNDWI_Norm)) / ((MBI_Norm) + (MNDWI_Norm))
    EMBI_Norm = MinMaxNorm(EMBI, MaskValid)

    invalid = (MBI_Norm == 0)
    MBI_Norm[invalid] = 1
    C = EMBI_Norm / MBI_Norm
    C[invalid] = 0

    # Derive the Bare Soil Index (BSI).
    BSI = ((SWIR1 + Red) - (NIR + Blue)) / ((SWIR1 + Red) + (NIR + Blue))
    MNDWI = HistCut(MNDWI, fillV, 6, [-1, 1])
    EMBI = HistCut(EMBI, fillV, 6, [-1, 1])
    BSI = HistCut(BSI, fillV, 6, [-1, 1])
    BSI_Norm = MinMaxNorm(BSI, MaskValid)
    # Deriving Eenhanced-BSI based on BSI and MNDWI.
    EBSI = ((BSI_Norm) - (MNDWI_Norm)) / ((BSI_Norm) + (MNDWI_Norm))
    MNDWI = HistCut(MNDWI, fillV, 6, [-1, 1])
    EMBI = HistCut(EMBI, fillV, 6, [-1, 1])
    BSI = HistCut(BSI, fillV, 6, [-1, 1])
    EBSI = HistCut(EBSI, fillV, 6, [-1, 1])

    # Derive SSF.
    SSF = C * (1 - EBSI)
    SSF_Norm = MinMaxNorm(SSF, MaskValid)

    # Modulation Factor (MF).
    MF = (Blue + Grn - NIR - SWIR1) / (Blue + Grn + NIR + SWIR1)
    MNDWI = HistCut(MNDWI, fillV, 6, [-1, 1])
    EMBI = HistCut(EMBI, fillV, 6, [-1, 1])
    BSI = HistCut(BSI, fillV, 6, [-1, 1])
    EBSI = HistCut(EBSI, fillV, 6, [-1, 1])

    MF = HistCut(MF, fillV, 6, [-1, 1])
    MF_Norm = MinMaxNorm(MF, MaskValid)

    # Derive ASI.
    ASI = AF_Norm * SSF_Norm * VSF_Norm * MF_Norm

    MNDWI = HistCut(MNDWI, fillV, 6, [-1, 1])
    EMBI = HistCut(EMBI, fillV, 6, [-1, 1])
    BSI = HistCut(BSI, fillV, 6, [-1, 1])
    EBSI = HistCut(EBSI, fillV, 6, [-1, 1])

    MF = HistCut(MF, fillV, 6, [-1, 1])
    ASI = HistCut(ASI, fillV, 6, [0, 1])

    # # The Artificial surface Factor (AF)
    # AF = (((SWIR1 + NIR) / 2) - Blue) / (((SWIR1 + NIR) / 2) + Blue) + 1

    # # %%%%% Vegetation Suppressing Factor (VSF).
    # # % EVI is better in general cases, but its adjustment for mountain shadows (with vegetation) is not as good as NDVI.
    # # EVI = G * ((NIR - Red) / (NIR + C1 * Red - C2 * Blue + L))
    # # https://en.wikipedia.org/wiki/Enhanced_vegetation_index
    # # Enhanced vegetation index
    # EVI = G * (NIR - Red) / (NIR + C1 * Red - C2 * Blue + L)
    # EVI = hist_cut(EVI, fillV)

    # # Normalized Difference Vegetation Index
    # # https://gisgeography.com/ndvi-normalized-difference-vegetation-index/
    # NDVI = (NIR - Red) / (NIR + Red)

    # # bare soil index
    # # https://www.geo.university/pages/blog?p=spectral-indices-with-multispectral-satellite-data#:~:text=Bare%20Soil%20Index%20(BSI)%20is,used%20in%20a%20normalized%20manner.
    # BSI = ((SWIR1 + Red) - (NIR + Blue)) / ((SWIR1 + Red) + (NIR + Blue))
    # BSI  = hist_cut( BSI, fillV, 6, minmax=[-1, 1])
    # BSI_Norm = minmax_norm(BSI, valid_mask)

    # # Modified Bare Soil Index
    # MBI = ((SWIR1 - SWIR2 - NIR) / (SWIR1 + SWIR2 + NIR)) + 0.5
    # MBI = hist_cut(MBI, fillV, minmax=[-0.5, 1.5])

    # # Note that BSI, MBI, and MNDWI need to be normalized when calculation EBSI
    # # and EMBI to avoid unmeaningful values caused by the negative values in
    # # BSI/MBI/MNDW
    # MNDWI = (Green - SWIR1) / (Green + SWIR1)
    # MNDWI  = hist_cut( MNDWI, fillV, 6, minmax=[-1, 1])
    # MNDWI_Norm = minmax_norm(MNDWI, valid_mask)

    # # EBSI/EMBI were designed based on BSI/MBI to enhance barren further,
    # # however, artificial surface still has relative high magnitude, thereby
    # # leading to the over-depressing of artificial surface.

    # # EBSI = (BSI - MNDWI) / (BSI + MNDWI)
    # EBSI = ((BSI_Norm) - (MNDWI_Norm)) / ((BSI_Norm) + (MNDWI_Norm));

    # EBSI  = hist_cut( EBSI, fillV, 6, minmax=[-1, 1])

    # EMBI = (MBI - MNDWI) / (MBI + MNDWI)
    # EMBI  = hist_cut( EMBI, fillV, 6, minmax=[-1, 1])

    # # Soil Depressing Factor
    # SDF = (EMBI + 1) / (MBI + 0.5) * (1 - EBSI)

    # # Vegetation Depressing Factor
    # # VDF is based on NDVI (Rouse et al. 1974) or EVI (Huete et al. 1997),
    # VDF = 1 - EVI if toplogical_effects else 1 - NDVI

    # # Therefore, we design a modulation factor to depress the dark bare land
    # # and enhance dark artificial surfaces simultaneous
    # MF = ((Blue + Green) - (NIR + SWIR1)) / ((Blue + Green) + (NIR + SWIR1)) + 1
    # MF  = hist_cut( MF, fillV, 6, minmax=[-1, 1])

    # ASI = (AF * SDF * VDF * MF) + 1
    # ASI  = hist_cut( ASI, fillV, 6, minmax=[0, 1])

    indexes = {
        'ASI': ASI,
        'MF': MF,
        # 'VDF': VDF,
        # 'SDF': SDF,
        'AF': AF,
        'EBSI': EBSI,
        'MNDWI': MNDWI,
        'MBI': MBI,
        'BSI': BSI,
        'BSI_Norm': BSI_Norm,
        'EVI': EVI,

        'EMBI': EMBI,

        'SSF_Norm': SSF_Norm,
        'AF_Norm': AF_Norm,
        'VSF_Norm': VSF_Norm,
        'MF_Norm': MF_Norm,
        # 'NDVI': NDVI,
    }

    return indexes


SPECIALIZED_BANDS = {
    'ASI',
    'MF',
    # 'VDF',
    # 'SDF',
    'AF',
    'EBSI',
    'MNDWI',
    'MBI',
    'BSI',
    'BSI_Norm',
    'EVI',

    'EMBI',

    'SSF_Norm',
    'AF_Norm',
    'VSF_Norm',
    'MF_Norm',
}


def specialized_index_bands2(delayed=None):
    r"""
    Ported from code from by (Yongquan Zhao on 26 April 2017)

    References:
        https://mail.google.com/mail/u/1/#chat/space/AAAAE5jpxTc

    Ignore:
        DVC_DPATH=$HOME/data/dvc-repos/smart_watch_dvc
        jq '.images[0].id' $DVC_DPATH/drop1-S2-L8-aligned/data.kwcoco.json
        kwcoco subset --src $DVC_DPATH/drop1-S2-L8-aligned/data.kwcoco.json --gids=2, --dst=./one_image_data/data.kwcoco.json --copy_assets=True

    TODO:
        - [ ] These depend on kwcoco, and kwgis should avoid depending on that.

    Ignore:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> from kwgis.utils.util_bands import *  # NOQA
        >>> from simple_dvc.registry import find_dvc_dpath
        >>> import kwcoco
        >>> dvc_dpath = find_dvc_dpath()
        >>> coco_fpath = dvc_dpath / 'Aligned-Drop4-2022-07-28-c20-TA1-S2-L8-ACC/data_vali.kwcoco.json'
        >>> dset = kwcoco.CocoDataset(coco_fpath)
        >>> #from kwgis.utils import kwcoco_extensions
        >>> gid = dset.images().compress([s == 'L8' for s in dset.images().get('sensor_coarse')]).objs[200]['id']
        >>> #gid = ub.peek(dset.index.imgs.keys())
        >>> coco_img = dset.coco_image(gid)
        >>> #print('coco_img.channels = {!r}'.format(coco_img.channels))
        >>> delayed = coco_img.imdelay(space='video', nodata_method='float')
        >>> symbolic = False
        >>> indexes = specialized_index_bands2(delayed)
        >>> import kwarray
        >>> #print(ub.urepr(ub.map_vals(kwarray.stats_dict, indexes), nl=1))
        >>> #import pandas as pd
        >>> #print(pd.DataFrame(ub.map_vals(kwarray.stats_dict, indexes)))
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> import kwimage
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(indexes) + 1)
        >>> kwplot.figure(fnum=3)
        >>> kwplot.imshow(value, title=key, pnum=pnum_(), cmap=None if key == 'MaskValid' else 'viridis', data_colorbar=True)
        >>> rgb = delayed.take_channels('red|green|blue').finalize()
        >>> rgb_canvas = kwimage.normalize_intensity(rgb)
        >>> kwplot.imshow(rgb_canvas, title='rgb', pnum=pnum_())
        >>> indexes['MaskValid'] = indexes['MaskValid'].astype(np.float32)
        >>> for key, value in indexes.items():
        >>>     value = value.astype(np.float32)
        >>>     #value = kwimage.normalize(value.astype(np.float32))
        >>>     kwplot.imshow(value, title=key, pnum=pnum_(), cmap=None if key == 'MaskValid' else 'viridis', data_colorbar=True)

    Ignore:
        >>> # xdoctest: +SKIP("Something is wrong with grabbing L8 images")
        >>> from kwgis.utils.util_bands import *  # NOQA
        >>> import kwgis
        >>> dset = kwgis.demo.demo_smart_raw_kwcoco()
        >>> coco_img = [img for img in dset.images().coco_images if img.get('sensor_coarse', None) == 'L8'][0]
        >>> delayed = coco_img.imdelay().crop((slice(4000, 5000), slice(4000, 5000)))
        >>> indexes = specialized_index_bands2(delayed)
        >>> # xdoctest: +REQUIRES(--show)
        >>> import kwplot
        >>> import kwimage
        >>> kwplot.autompl()
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(indexes))
        >>> kwplot.figure(fnum=3)
        >>> indexes['MaskValid'] = indexes['MaskValid'].astype(np.float32)
        >>> for key, value in indexes.items():
        >>>     value = value.astype(np.float32)
        >>>     #value = kwimage.normalize(value.astype(np.float32))
        >>>     kwplot.imshow(value, title=key, pnum=pnum_(), cmap=None if key == 'MaskValid' else 'viridis', data_colorbar=True)


    Ignore:
        >>> delayed = coco_img.imdelay()
        >>> rgbir123 = delayed.take_channels('blue|green|red|swir16|swir22|nir')
        >>> chw = rgbir123.finalize().transpose(2, 0, 1).astype(np.float32)
        >>> Blue, Green, Red, SWIR1, SWIR2, NIR = chw
        >>> df = pd.DataFrame(chw.reshape(6, -1).T, columns=['blue', 'red', 'green', 'swir1', 'swir2', 'nir'])
        >>> df = df.melt()
        >>> import kwplot
        >>> sns = kwplot.autosns()
        >>> palette = {
        >>>     'red': 'red', 'blue': 'blue', 'green': 'green',
        >>>     'swir1': 'purple', 'swir2': 'orange', 'nir': 'pink',
        >>> }
        >>> kwplot.figure(fnum=2)
        >>> sns.histplot(data=df, x='value', hue='variable', palette=palette)
        >>> kwplot.figure(fnum=4)
        >>> raw = {
        >>>     'Red': Red,
        >>>     'Blue': Blue,
        >>>     'Green': Green,
        >>>     'NIR': NIR,
        >>>     'SWIR1': SWIR1,
        >>>     'SWIR2': SWIR2,
        >>> }
        >>> pnum_ = kwplot.PlotNums(nSubplots=len(raw))
        >>> for key, value in raw.items():
        >>>     #value = kwimage.normalize(value.astype(np.float32))
        >>>     value = value.astype(np.float32)
        >>>     kwplot.imshow(value, title=key, pnum=pnum_(), cmap='viridis', data_colorbar=True)
    """
    import numpy as np
    # Raw bands
    # if symbolic:
    #     # Sympy can help explore different forms of these equations.
    #     import sympy as sym
    #     Blue, Green, Red, SWIR1, SWIR2, NIR = sym.symbols(
    #         'Blue, Green, Red, SWIR1, SWIR2, NIR')
    # else:
    rgbir123 = delayed.take_channels('blue|green|red|swir16|swir22|nir')
    chw = rgbir123.finalize().transpose(2, 0, 1).astype(np.float32)

    # Artificial Surface Index (ASI) is designed based the surface reflectance imagery of Landsat 8.
    # The data value range of Landsat surface reflectance [0, 1] is
    # transformed to [0, 1*Scale].
    # Scale = 10000
    Scale = 15000

    hack_norm = 0
    if hack_norm:
        # How do we apply this to S2?
        import kwimage
        chw = kwimage.normalize_intensity(chw)
        Scale = 1

    Blue, Green, Red, SWIR1, SWIR2, NIR = chw

    def hist_cut(band, fill_value=0, k=3, minmax='std'):
        if minmax == 'std':
            mean = band.mean()
            std = band.std()
            low_val = (mean - k * std)  # Corrected.
            high_val = (mean + k * std)
        else:
            low_val, high_val = minmax
        band = band.clip(low_val, high_val)
        return band

    def minmax_norm(band, mask):
        max_val = band[mask].max()
        min_val = band[mask].min()
        extent = max_val - min_val
        if extent != 0:
            shifted = band - min_val
            scaled = shifted / extent
            band[mask] = scaled[mask]
        return band

    fillV = 0

    # Surface reflectance should be within [0, 1*Scale]
    max_vals = np.maximum.reduce([Blue, Green, Red, NIR, SWIR1, SWIR2])
    min_vals = np.minimum.reduce([Blue, Green, Red, NIR, SWIR1, SWIR2])
    MaskValid = (0 < min_vals) & (max_vals < Scale)
    # MaskValid = MaskValid.astype(np.uint8)

    # Artificial surface Factor (AF).
    AF = (NIR - Blue) / (NIR + Blue)
    AF = hist_cut(AF, fillV, 6, [-1, 1])
    AF_Norm = minmax_norm(AF, MaskValid)

    # Vegetation Suppressing Factor (VSF).
    # Modified Soil Adjusted Vegetation Index (MSAVI).
    MSAVI = (2 * NIR + 1 * Scale -
             ((2 * NIR + 1 * Scale)**2 - 8 * (NIR - Red)) ** 0.5) / 2
    MSAVI = hist_cut(MSAVI, fillV, 6, [-1, 1])
    NDVI = (NIR - Red) / (NIR + Red)
    NDVI = hist_cut(NDVI, fillV, 6, [-1, 1])
    VSF = 1 - MSAVI * NDVI
    VSF_Norm = minmax_norm(VSF, MaskValid)

    # Soil Suppressing Factor (SSF).
    # Derive the Modified Bare soil Index (MBI).
    MBI = (SWIR1 - SWIR2 - NIR) / (SWIR1 + SWIR2 + NIR) + 0.5
    MBI = hist_cut(MBI, fillV, 6, [-0.5, 1.5])
    # Deriving Enhanced-MBI based on MBI and MNDWI.
    MNDWI = (Green - SWIR1) / (Green + SWIR1)
    MNDWI = hist_cut(MNDWI, fillV, 6, [-1, 1])
    EMBI = ((MBI + 0.5) - (MNDWI + 1)) / ((MBI + 0.5) + (MNDWI + 1))
    EMBI = hist_cut(EMBI, fillV, 6, [-1, 1])
    # Derive SSF.
    SSF = (1 - EMBI)
    SSF_Norm = minmax_norm(SSF, MaskValid)

    # Modulation Factor (MF).
    MF = (Blue + Green - NIR - SWIR1) / (Blue + Green + NIR + SWIR1)
    MF = hist_cut(MF, fillV, 6, [-1, 1])
    MF_Norm = minmax_norm(MF, MaskValid)

    # Derive Artificial Surface Index (ASI).
    ASI = AF_Norm * SSF_Norm * VSF_Norm * MF_Norm
    ASI = hist_cut(ASI, fillV, 6, [0, 1])
    ASI = ASI * MaskValid

    indexes = {
        'MaskValid': MaskValid,
        'ASI': ASI,

        'MF': MF,
        'MF_Norm': MF_Norm,

        'AF': AF,
        'AF_Norm': AF_Norm,

        'SSF_Norm': SSF_Norm,
        'VSF_Norm': VSF_Norm,

        'MSAVI': MSAVI,
        'NDVI': NDVI,
        'MNDWI': MNDWI,
        'MBI': MBI,

        'EMBI': EMBI,
    }

    return indexes



"""
Module for access to simple data for demo and testing purposes.
"""

__devnotes__ = """
mkinit -m kwgis.demo --lazy -w
"""


def lazy_import(module_name, submodules, submod_attrs):
    import importlib
    import os
    name_to_submod = {
        func: mod for mod, funcs in submod_attrs.items()
        for func in funcs
    }

    def __getattr__(name):
        if name in submodules:
            attr = importlib.import_module(
                '{module_name}.{name}'.format(
                    module_name=module_name, name=name)
            )
        elif name in name_to_submod:
            submodname = name_to_submod[name]
            module = importlib.import_module(
                '{module_name}.{submodname}'.format(
                    module_name=module_name, submodname=submodname)
            )
            attr = getattr(module, name)
        else:
            raise AttributeError(
                'No {module_name} attribute {name}'.format(
                    module_name=module_name, name=name))
        globals()[name] = attr
        return attr

    if os.environ.get('EAGER_IMPORT', ''):
        for name in name_to_submod.values():
            __getattr__(name)

        for attrs in submod_attrs.values():
            for attr in attrs:
                __getattr__(attr)
    return __getattr__


__getattr__ = lazy_import(
    __name__,
    submodules={
        'dummy_demodata',
        'landsat_demodata',
        'nitf_demodata',
        'sentinel2_demodata',
        'smart_kwcoco_demodata',
        'stac_demo',
    },
    submod_attrs={
        'dummy_demodata': [
            'dummy_rpc_geotiff_fpath',
        ],
        'landsat_demodata': [
            'grab_landsat_product',
        ],
        'nitf_demodata': [
            'DEFAULT_KEY',
            'grab_nitf_fpath',
        ],
        'sentinel2_demodata': [
            'grab_sentinel2_product',
        ],
        'smart_kwcoco_demodata': [
            
            'demo_kwcoco_multisensor',
            'demo_kwcoco_with_heatmaps',
            'hack_seed_geometadata_in_dset',
        ],
        'stac_demo': [
            'demo',
        ],
    },
)


def __dir__():
    return __all__


__all__ = ['DEFAULT_KEY', 'demo', 'demo_kwcoco_multisensor',
           'demo_kwcoco_with_heatmaps', 'dummy_demodata',
           'dummy_rpc_geotiff_fpath', 'grab_landsat_product',
           'grab_nitf_fpath', 'grab_sentinel2_product',
           'hack_seed_geometadata_in_dset', 'landsat_demodata',
           'nitf_demodata', 'sentinel2_demodata', 'smart_kwcoco_demodata',
           'stac_demo']



"""
Functions to perterb or "jitter" truth that gradually degrade the site models
in specified ways. This can be used to check the scoring metrics at various
levels of accuracy / inaccuracy.
"""
import numpy as np
import geojson
import kwarray
import kwimage
import ubelt as ub
import geopandas as gpd
from kwgis.demo.metrics_demo import demo_truth
from kwgis.demo.metrics_demo import demo_utils


def perterb_site_model(sites, rng=None, **kwargs):
    """
    Given a true site models from a region, perterb them to make "demo"
    predicted site models.

    Args:
        sites (List[geojson.FeatureCollection]): geojson site observations
        rng : random seed or generator
        **kwargs: factors to control per-site perterbations. See
            :func:`perterb_single_site_model` for available options.

    Returns:
        List[geojson.FeatureCollection]: modified site models

    Example:
        >>> from kwgis.demo.metrics_demo.site_perterbing import *  # NOQA
        >>> _, sites, _ = demo_truth.random_region_model(rng=12345)
        >>> pred_sites1 = perterb_site_model(sites, noise=1, rng=34567)
        >>> pred_sites2 = perterb_site_model(sites, noise=0, rng=34567)
    """
    rng = kwarray.ensure_rng(rng)

    # Shuffle our site models so the indices don't correspond.
    templates = sites.copy()
    rng.shuffle(templates)

    # TODO: drop sites to simulate false negatives and add new random sites to
    # generate false positives.

    pred_sites = []
    for idx, site in enumerate(templates):
        try:
            pred_site = perterb_single_site_model(site, idx=idx, rng=rng, **kwargs)
        except IndexError:
            continue  # drop the site
        pred_sites.append(pred_site)
    return pred_sites


class PerterbModel:
    """
    TODO:
        consume the functionality of perterb_single_site_model

    Example:
        >>> from kwgis.demo.metrics_demo.site_perterbing import *  # NOQA
        >>> self = PerterbModel()

    Example:
        >>> from kwgis.demo.metrics_demo.site_perterbing import *  # NOQA
        >>> from kwgis.demo.metrics_demo import demo_truth
        >>> region_id = 'DR_0042'
        >>> site_id = 'DR_0042_9001'
        >>> rng = kwarray.ensure_rng(4222)
        >>> region_corners = kwimage.Boxes([[5, 7, 11, 13]], 'xywh').to_ltrb().corners()
        >>> observables = demo_truth.random_observables(15, rng=rng)
        >>> p_observe = 0.5
        >>> site_summary, site = demo_truth.random_site_model(
        >>>     region_id, site_id, region_corners, observables,
        >>>     p_observe=p_observe, p_transition=0.4,  rng=rng)
        >>> kwargs = dict(noise=10, drop_limit=0.1)
        >>> self = PerterbModel(**kwargs)
        >>> pred_site = self.perterb_single_site(site)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> import geopandas as gpd
        >>> # Draw our true and perterbed site model
        >>> orig_site_gdf = gpd.GeoDataFrame.from_features(site)
        >>> pred_site_gdf = gpd.GeoDataFrame.from_features(pred_site)
        >>> total_poly1 = orig_site_gdf['geometry'].unary_union
        >>> total_poly2 = pred_site_gdf['geometry'].unary_union
        >>> total_poly = total_poly1.union(total_poly2)
        >>> minx, miny, maxx, maxy = total_poly.bounds
        >>> fig = kwplot.figure(doclf=True, fnum=1)
        >>> phases = demo_truth.SiteModelGenerator.SITE_PHASES
        >>> phase_to_color = ub.udict(ub.dzip(phases, kwimage.Color.distinct(len(phases))))
        >>> pnum_ = kwplot.PlotNums(nRows=4, nSubplots=len(observables) + 1)
        >>> boundary_colors = ub.udict({'true': 'limegreen', 'pred': 'dodgerblue'}).map_values(lambda c: kwimage.Color(c).as01())
        >>> for obs in observables:
        >>>     title = obs['datetime'].isoformat()
        >>>     trues = orig_site_gdf[orig_site_gdf['observation_date'] == obs['datetime'].date().isoformat()]
        >>>     preds = pred_site_gdf[pred_site_gdf['observation_date'] == obs['datetime'].date().isoformat()]
        >>>     fig = kwplot.figure(pnum=pnum_(), title=title)
        >>>     ax = fig.gca()
        >>>     if len(trues):
        >>>         colors = list(phase_to_color.take(trues['current_phase']))
        >>>         trues.plot(ax=ax, alpha=0.5, color=colors)
        >>>         trues.boundary.plot(ax=ax, alpha=0.5, color=boundary_colors['true'], linewidth=3)
        >>>         trues.centroid.plot(ax=ax, alpha=0.5, color=boundary_colors['true'])
        >>>     if len(preds):
        >>>         colors = list(phase_to_color.take(preds['current_phase']))
        >>>         preds.plot(ax=ax, alpha=0.5, color=colors)
        >>>         preds.boundary.plot(ax=ax, alpha=0.5, color=boundary_colors['pred'], linewidth=3)
        >>>         preds.centroid.plot(ax=ax, alpha=0.5, color=boundary_colors['pred'])
        >>>     ax.set_xlim(minx, maxx)
        >>>     ax.set_ylim(miny, maxy)
        >>> legend1 = kwimage.draw_header_text(kwplot.make_legend_img(phase_to_color), 'Face Key', fit=True)
        >>> legend2 = kwimage.draw_header_text(kwplot.make_legend_img(boundary_colors), 'Border Key', fit=True)
        >>> legend = kwimage.stack_images([legend1, legend2], axis=0, resize='smaller')
        >>> kwplot.imshow(legend, pnum=pnum_())
        >>> kwplot.show_if_requested()
    """

    def __init__(self, noise=0.0, performer_id='alice', rng=None, **kwargs):
        self.rng = rng
        self.performer_id = performer_id

        default_noise = ub.udict({
            'warp_noise': 1.0,
            'label_noise': 1.0,
            'drop_noise': 0.5,
        })
        default_limits = ub.udict({
            'drop_limit': 0.5,
            'scale_limit': 0.3,
            'theta_limit': np.pi / 16,
        })
        kwargs = ub.udict(kwargs)

        limits = default_limits | (kwargs & default_limits)
        kwargs -= limits
        limits = limits.map_keys(lambda k: k.split('_limit')[0])

        noises = default_noise | (kwargs & default_noise)
        kwargs -= noises
        noises = noises.map_keys(lambda k: k.split('_noise')[0])

        self.limits = limits
        self.noises = noises
        self.noise = noise

        if len(kwargs) > 0:
            raise ValueError('unknown kwargs={}'.format(kwargs))

        self.distris = None
        self._setup_distributions()

    def _setup_distributions(self):
        from kwarray import distributions as distri
        rng = self.rng
        noises = self.noises
        limits = self.limits
        max_scale = 1 + limits['scale'] * (noises['warp'] * self.noise)
        max_theta = 0 + limits['theta'] * (noises['warp'] * self.noise)
        p_randomize = max(0, np.tanh(noises['label'] * self.noise))
        p_drop = min(max(0, np.tanh(noises['drop'] * self.noise)), limits['drop'])

        distris = {}
        distris['scale'] = distri.Uniform(1, max_scale, rng=rng) ** distri.CategoryUniform([-1, 1])
        distris['theta'] = distri.Uniform(0, max_theta, rng=rng) * distri.CategoryUniform([-1, 1])
        distris['phase'] = distri.CategoryUniform(demo_truth.SiteModelGenerator.SITE_PHASES, rng=rng)
        distris['randomize_phase'] = distri.Bernoulli(p_randomize, rng=rng)
        distris['drop_frame'] = distri.Bernoulli(p_drop, rng=rng)
        self.distris = distris

    def perterb_single_site(self, site, idx=None):
        true_site_summary = site["features"][0]["properties"]
        mgrs_code = true_site_summary["mgrs"]

        # TODO: add the ability to pass in observables to add temporal noise

        # TODO: could be more complex here.
        valid_status_codes = [
            "system_proposed",
            "system_confirmed",
            "system_rejected",
        ]
        status = valid_status_codes[1]

        # Find the true site header
        orig_site_header = None
        orig_observations = []
        for feat in site["features"]:
            if feat['properties']['type'] == 'site':
                if orig_site_header is not None:
                    raise ValueError('More than one site header')
                orig_site_header = feat
            elif feat['properties']['type'] == 'observation':
                orig_observations.append(feat)
            else:
                raise TypeError('Unknown feature type, should be site or observation')
        if orig_site_header is None:
            raise ValueError(
                'A site feature collection does not have a summary header '
                'with type=site in its properties')

        region_id = orig_site_header['properties']['region_id']
        if idx is None:
            site_id = orig_site_header['properties']['site_id']
        else:
            site_id = f"{region_id}_{idx:03d}"

        orig_obs_crs84 = gpd.GeoDataFrame.from_features(
            orig_observations, crs=demo_utils.get_crs84())

        # Do any shape warping in the local UTM space.
        orig_obs_utm = demo_utils.project_gdf_to_local_utm(orig_obs_crs84)
        orig_geoms_utm = orig_obs_utm['geometry']

        new_geoms_utm = []
        new_properties = []
        for feat, geom_utm in zip(orig_observations, orig_geoms_utm):

            if self.distris['drop_frame'].sample():
                continue

            orig_props = ub.udict(feat["properties"])
            poly = kwimage.MultiPolygon.coerce(geom_utm)

            xy = [d[0] for d in poly.to_shapely().centroid.xy]
            rand_aff_params = {
                'scale': self.distris['scale'].sample(),
                'theta': self.distris['theta'].sample(),
            }
            aff = kwimage.Affine.affine(**rand_aff_params, about=xy)
            new_poly = poly.warp(aff)

            current_phase = orig_props["current_phase"]

            if self.distris['randomize_phase'].sample():
                current_phase = self.distris['phase'].sample()

            new_props = orig_props | {
                "id": site_id,
                "current_phase": current_phase,
                "is_site_boundary": True,
                "originator": self.performer_id,
                "validated": "False",
            }
            new_properties.append(new_props)
            new_geoms_utm.append(new_poly.to_shapely())

        # Convert back to CRS84
        new_geoms_utm = gpd.GeoDataFrame(geometry=new_geoms_utm, crs=orig_obs_utm.crs)
        new_summary_utm = gpd.GeoDataFrame(geometry=[new_geoms_utm.unary_union], crs=orig_obs_utm.crs)

        new_summary_crs84 = new_summary_utm.to_crs(orig_obs_crs84.crs)['geometry'].iloc[0]
        new_geoms_crs84 = new_geoms_utm.to_crs(orig_obs_crs84.crs)['geometry']

        # Build the geojson features
        new_observations = []
        for new_props, new_geom in zip(new_properties, new_geoms_crs84):
            new_poly = kwimage.MultiPolygon.coerce(new_geom)
            new = geojson.Feature(
                geometry=new_poly.to_geojson(),
                properties=new_props,
            )
            new_observations.append(new)

        new_site_header = demo_truth.make_site_summary(
            new_observations, mgrs_code, site_id, status,
            summary_geom=new_summary_crs84)
        new_site_header["properties"].update(
            {
                "type": "site",
                "site_id": site_id,
                "region_id": region_id,
            }
        )
        pred_site = geojson.FeatureCollection([new_site_header] + new_observations)
        return pred_site


def perterb_single_site_model(site, noise=0.0, warp_noise=1.0,
                              performer_id='alice', idx=None, rng=None, **kwargs):
    """
    Given a true site model, perterb it to make a "demo" predicted site model.

    Args:
        site (geojson.FeatureCollection): geojson site observations

        noise (float):
            Magnitude of all noise. This factor modulates all other noise
            values. Setting to 0 disables all perterbation. Typically 1
            is the largest "reasonable" value. Setting any noise magnitude
            higher than 1 may result in pathological perterbations.

        warp_noise (float):
            magnitude of random warp that we will use to perterb the boundary
            of the true site polygons.

        performer_id (str):
            used in the originator property

        idx (int): the index of this new site model.
            If unspecified, the same site id is used in truth and pred.

        rng : random seed or generator

        **kwargs : other PerterbModel params

    Returns:
        geojson.FeatureCollection: modified site model

    Example:
        >>> # Demo case with a lot of noise
        >>> from kwgis.demo.metrics_demo.site_perterbing import *  # NOQA
        >>> rng = kwarray.ensure_rng(43240830)
        >>> _, sites, _ = demo_truth.random_region_model(num_observations=20, rng=rng)
        >>> site = sites[0]
        >>> pred_site = perterb_single_site_model(site, noise=1.0, rng=rng)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> # Draw our true and perterbed site model
        >>> orig_site_gdf = gpd.GeoDataFrame.from_features(site)
        >>> pred_site_gdf = gpd.GeoDataFrame.from_features(pred_site)
        >>> fig = kwplot.figure(doclf=True, fnum=1)
        >>> ax = fig.gca()
        >>> total_poly = pred_site_gdf['geometry'].unary_union.union(orig_site_gdf['geometry'].unary_union)
        >>> minx, miny, maxx, maxy = total_poly.bounds
        >>> ax.set_xlim(minx, maxx)
        >>> ax.set_ylim(miny, maxy)
        >>> pred_site_gdf.plot(ax=ax, alpha=0.5, color='orange', edgecolor='purple')
        >>> orig_site_gdf.plot(ax=ax, alpha=0.5, color='limegreen', edgecolor='black')
        >>> orig_site_gdf.centroid.plot(ax=ax, alpha=0.5, color='green')
        >>> pred_site_gdf.centroid.plot(ax=ax, alpha=0.5, color='red')
        >>> kwplot.show_if_requested()

    Example:
        >>> # Demo case with almost zero noise
        >>> from kwgis.demo.metrics_demo.site_perterbing import *  # NOQA
        >>> rng = kwarray.ensure_rng(43240830)
        >>> _, sites, _ = demo_truth.random_region_model(num_observations=20, rng=rng)
        >>> site = sites[0]
        >>> pred_site = perterb_single_site_model(site, noise=1e-2, rng=rng)
        >>> # xdoctest: +REQUIRES(--show)
        >>> # xdoctest: +REQUIRES(module:kwplot)
        >>> import kwplot
        >>> kwplot.autompl()
        >>> # Draw our true and perterbed site model
        >>> orig_site_gdf = gpd.GeoDataFrame.from_features(site)
        >>> pred_site_gdf = gpd.GeoDataFrame.from_features(pred_site)
        >>> fig = kwplot.figure(doclf=True, fnum=2)
        >>> ax = fig.gca()
        >>> total_poly = pred_site_gdf['geometry'].unary_union.union(orig_site_gdf['geometry'].unary_union)
        >>> minx, miny, maxx, maxy = total_poly.bounds
        >>> ax.set_xlim(minx, maxx)
        >>> ax.set_ylim(miny, maxy)
        >>> pred_site_gdf.plot(ax=ax, alpha=0.5, color='orange', edgecolor='purple')
        >>> orig_site_gdf.plot(ax=ax, alpha=0.5, color='limegreen', edgecolor='black')
        >>> orig_site_gdf.centroid.plot(ax=ax, alpha=0.5, color='green')
        >>> pred_site_gdf.centroid.plot(ax=ax, alpha=0.5, color='red')
        >>> kwplot.show_if_requested()
    """

    perterber = PerterbModel(noise=noise, warp_noise=warp_noise,
                             performer_id=performer_id, rng=rng, **kwargs)
    pred_site = perterber.perterb_single_site(site, idx=idx)
    return pred_site



"""
A reimplementation of the retry library [RetryLib]_ with some minor changes,
mainly logging default to a print statement rather than an actual logger.

References:
    .. [RetryLib] https://github.com/invl/retry
"""
import time
import random


class DummyLogger:
    def warning(self, msg, *args):
        print(msg % args)


class Retry:
    """
    Reimplementation of the retry internals
    """
    def __init__(self, exceptions=Exception, tries=-1, delay=0, backoff=1, max_delay=None, jitter=0, logger=None):
        if tries <= 0:
            raise NotImplementedError('tries <= 0 is not supported')
        self.tries = tries
        self.delay = delay
        self.exceptions = exceptions
        self.backoff = backoff
        if logger is None:
            logger = DummyLogger()
        self.logger = logger
        self.jitter = jitter
        if max_delay is None:
            max_delay = float('inf')
        self.max_delay = max_delay

    def __call__(self, func, *args, **kwargs):
        current_delay = self.delay
        for try_num in range(1, self.tries + 1):
            try:
                return func(*args, **kwargs)
            except self.exceptions as ex:
                if try_num >= self.tries:
                    self.logger.warning(f'Error on try {try_num}/{self.tries}. ex={ex!r}, giving up.')
                    raise
                else:
                    self.logger.warning(f'Error on try {try_num}/{self.tries}. ex={ex!r}, retry after delay={current_delay:0.2f} seconds')

            if isinstance(current_delay, tuple):
                current_delay += random.uniform(*self.jitter)
            else:
                current_delay += self.jitter

            time.sleep(current_delay)
            current_delay *= self.backoff
            current_delay = min(current_delay, self.max_delay)


def retry_call(f, fargs=None, fkwargs=None, exceptions=Exception, tries=-1,
               delay=0, max_delay=None, backoff=1, jitter=0, logger=None):
    """
    Retry API compatable

    CommandLine:
        xdoctest -m kwgis.utils.util_retry retry_call

    Example:
        >>> from kwgis.utils.util_retry import retry_call
        >>> _context = {'attempt': 0}
        >>> def f():
        >>>     _context['attempt'] += 1
        >>>     if _context['attempt'] <= 5:
        >>>         raise Exception
        >>>     return 1
        >>> import pytest
        >>> with pytest.raises(Exception):
        ...     result = retry_call(f, tries=2, delay=0.01)
        >>> result = retry_call(f, tries=4, delay=0.01)
    """
    retry_obj = Retry(
        tries=tries, delay=delay, backoff=backoff, exceptions=exceptions,
        jitter=jitter, max_delay=max_delay, logger=logger)
    fargs = fargs or []
    fkwargs = fkwargs or {}
    return retry_obj(f, *fargs, **fkwargs)
